import os
import pytest
from pydm import Display

# The path to the .ui file used in these tests
test_ui_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "test_data", "test.ui")


def test_ui_filename_arg(qtbot):
    """If you supply a valid filename argument, you shouldn't get any exceptions."""
    my_display = Display(parent=None, ui_filename=test_ui_path)
    qtbot.addWidget(my_display)


def test_reimplemented_ui_filename(qtbot):
    """If you reimplement ui_filename and return a valid filename, you
    shouldn't get any exceptions."""
    class TestDisplay(Display):
        def ui_filename(self):
            return test_ui_path
    my_display = TestDisplay(parent=None)
    qtbot.addWidget(my_display)


def test_nonexistant_ui_file_raises(qtbot):
    with pytest.raises(IOError):
        my_display = Display(parent=None, ui_filename="this_doesnt_exist.ui")

    class TestDisplay(Display):
        def ui_filename(self):
            return "this_doesnt_exist.ui"

    with pytest.raises(IOError):
        my_display = TestDisplay(parent=None)



import os
import json
import functools
import webbrowser

from qtpy import QtWidgets, QtCore, QtDesigner
from ..utilities.iconfont import IconFont


class RulesEditor(QtWidgets.QDialog):
    """
    QDialog for user-friendly editing of the Rules in a widget inside the Qt
    Designer.

    Parameters
    ----------
    widget : PyDMWidget
        The widget which we want to edit the `rules` property.
    """

    def __init__(self, widget, parent=None):
        super(RulesEditor, self).__init__(parent)

        self.widget = widget
        self.lst_rule_item = None
        self.loading_data = True

        self.available_properties = widget.RULE_PROPERTIES
        self.default_property = widget.DEFAULT_RULE_PROPERTY

        self.setup_ui()

        try:
            self.rules = json.loads(widget.rules)
        except:
            self.rules = []

        for ac in self.rules:
            self.lst_rules.addItem(ac.get("name", ''))

    def setup_ui(self):
        """
        Create the required UI elements for the form.

        Returns
        -------
        None
        """
        iconfont = IconFont()

        self.setWindowTitle("PyDM Widget Rules Editor")
        vlayout = QtWidgets.QVBoxLayout()
        vlayout.setContentsMargins(5, 5, 5, 5)
        vlayout.setSpacing(5)
        self.setLayout(vlayout)

        hlayout = QtWidgets.QHBoxLayout()
        hlayout.setContentsMargins(0, 0, 0, 0)
        hlayout.setSpacing(5)
        vlayout.addLayout(hlayout)

        # Creating the widgets for the String List and
        # buttons to add and remove actions
        list_frame = QtWidgets.QFrame(parent=self)
        list_frame.setMinimumHeight(300)
        list_frame.setMinimumWidth(240)
        list_frame.setLineWidth(1)
        list_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        list_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        lf_layout = QtWidgets.QVBoxLayout()
        list_frame.setLayout(lf_layout)

        lf_btn_layout = QtWidgets.QHBoxLayout()
        lf_btn_layout.setContentsMargins(0, 0, 0, 0)
        lf_btn_layout.setSpacing(5)

        self.btn_add_rule = QtWidgets.QPushButton(parent=self)
        self.btn_add_rule.setAutoDefault(False)
        self.btn_add_rule.setDefault(False)
        self.btn_add_rule.setText("Add Rule")
        self.btn_add_rule.clicked.connect(self.add_rule)

        self.btn_del_rule = QtWidgets.QPushButton(parent=self)
        self.btn_del_rule.setAutoDefault(False)
        self.btn_del_rule.setDefault(False)
        self.btn_del_rule.setText("Remove Rule")
        self.btn_del_rule.clicked.connect(self.del_rule)

        lf_btn_layout.addWidget(self.btn_add_rule)
        lf_btn_layout.addWidget(self.btn_del_rule)

        lf_layout.addLayout(lf_btn_layout)

        self.lst_rules = QtWidgets.QListWidget()
        self.lst_rules.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding))
        self.lst_rules.itemSelectionChanged.connect(self.load_from_list)
        lf_layout.addWidget(self.lst_rules)

        hlayout.addWidget(list_frame)

        buttons_layout = QtWidgets.QHBoxLayout()
        save_btn = QtWidgets.QPushButton("Save", parent=self)
        save_btn.setAutoDefault(False)
        save_btn.setDefault(False)
        save_btn.clicked.connect(self.saveChanges)
        cancel_btn = QtWidgets.QPushButton("Cancel", parent=self)
        cancel_btn.setAutoDefault(False)
        cancel_btn.setDefault(False)
        cancel_btn.clicked.connect(self.cancelChanges)
        buttons_layout.addStretch()
        buttons_layout.addWidget(cancel_btn)
        buttons_layout.addWidget(save_btn)

        vlayout.addLayout(buttons_layout)

        # Creating the widgets that we will use to compose the
        # rule parameters
        self.frm_edit = QtWidgets.QFrame()
        self.frm_edit.setEnabled(False)
        self.frm_edit.setLineWidth(1)
        self.frm_edit.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frm_edit.setFrameShape(QtWidgets.QFrame.StyledPanel)

        frm_edit_layout = QtWidgets.QVBoxLayout()
        self.frm_edit.setLayout(frm_edit_layout)

        hlayout.addWidget(self.frm_edit)

        edit_name_layout = QtWidgets.QFormLayout()
        edit_name_layout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        lbl_name = QtWidgets.QLabel("Rule Name:")
        self.txt_name = QtWidgets.QLineEdit()
        self.txt_name.editingFinished.connect(self.name_changed)
        edit_name_layout.addRow(lbl_name, self.txt_name)
        lbl_property = QtWidgets.QLabel("Property:")
        self.cmb_property = QtWidgets.QComboBox()
        for name, prop in self.available_properties.items():
            self.cmb_property.addItem(name, prop)
        edit_name_layout.addRow(lbl_property, self.cmb_property)

        frm_edit_layout.addLayout(edit_name_layout)

        btn_add_remove_layout = QtWidgets.QHBoxLayout()
        self.btn_add_channel = QtWidgets.QPushButton()
        self.btn_add_channel.setAutoDefault(False)
        self.btn_add_channel.setDefault(False)
        self.btn_add_channel.setText("Add Channel")
        self.btn_add_channel.setIconSize(QtCore.QSize(16, 16))
        self.btn_add_channel.setIcon(iconfont.icon("plus-circle"))
        self.btn_add_channel.clicked.connect(self.add_channel)
        self.btn_del_channel = QtWidgets.QPushButton()
        self.btn_del_channel.setAutoDefault(False)
        self.btn_del_channel.setDefault(False)
        self.btn_del_channel.setText("Remove Channel")
        self.btn_del_channel.setIconSize(QtCore.QSize(16, 16))
        self.btn_del_channel.setIcon(iconfont.icon("minus-circle"))
        self.btn_del_channel.clicked.connect(self.del_channel)
        btn_add_remove_layout.addWidget(self.btn_add_channel)
        btn_add_remove_layout.addWidget(self.btn_del_channel)

        frm_edit_layout.addLayout(btn_add_remove_layout)

        self.tbl_channels = QtWidgets.QTableWidget()
        self.tbl_channels.setMinimumWidth(350)
        self.tbl_channels.setShowGrid(True)
        self.tbl_channels.setCornerButtonEnabled(False)
        self.tbl_channels.model().dataChanged.connect(self.tbl_channels_changed)
        headers = ["Channel", "Trigger?", "Enum?"]
        self.tbl_channels.setColumnCount(len(headers))
        self.tbl_channels.setHorizontalHeaderLabels(headers)
        header = self.tbl_channels.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(2, QtWidgets.QHeaderView.ResizeToContents)

        frm_edit_layout.addWidget(self.tbl_channels)

        expression_layout = QtWidgets.QFormLayout()
        expression_layout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        lbl_expected = QtWidgets.QLabel("Expected Type:")
        self.lbl_expected_type = QtWidgets.QLabel(parent=self)
        # self.lbl_expected_type.setText("")
        self.lbl_expected_type.setStyleSheet(
            "color: rgb(0, 128, 255); font-weight: bold;")
        expression_layout.addRow(lbl_expected, self.lbl_expected_type)

        lbl_initial = QtWidgets.QLabel("Initial Value:")
        self.txt_initial_value = QtWidgets.QLineEdit()
        self.txt_initial_value.editingFinished.connect(self.initial_value_changed)
        expression_layout.addRow(lbl_initial, self.txt_initial_value)

        lbl_expression = QtWidgets.QLabel("Expression:")
        expr_help_layout = QtWidgets.QHBoxLayout()
        self.txt_expression = QtWidgets.QLineEdit()
        self.txt_expression.editingFinished.connect(self.expression_changed)
        expr_help_layout.addWidget(self.txt_expression)
        self.btn_help = QtWidgets.QPushButton()
        self.btn_help.setAutoDefault(False)
        self.btn_help.setDefault(False)
        self.btn_help.setText("Help")
        self.btn_help.setStyleSheet("background-color: rgb(176, 227, 255);")
        self.btn_help.clicked.connect(functools.partial(self.open_help, open=True))
        expr_help_layout.addWidget(self.btn_help)
        expression_layout.addRow(lbl_expression, expr_help_layout)

        self.cmb_property.currentIndexChanged.connect(self.property_changed)
        self.cmb_property.setCurrentText(self.default_property)

        frm_edit_layout.addLayout(expression_layout)

    def clear_form(self):
        """Clear the form and reset the fields."""
        self.loading_data = True
        self.lst_rule_item = None
        self.txt_name.setText("")
        self.cmb_property.setCurrentIndex(-1)
        self.tbl_channels.clearContents()
        self.txt_initial_value.setText("")
        self.txt_expression.setText("")
        self.frm_edit.setEnabled(False)
        self.loading_data = False

    def load_from_list(self):
        """
        Load an entry from the list into the editing form.

        Returns
        -------
        None
        """
        item = self.lst_rules.currentItem()
        idx = self.lst_rules.indexFromItem(item).row()

        if idx < 0:
            return

        self.loading_data = True
        self.lst_rule_item = item
        data = self.rules[idx]
        self.txt_name.setText(data.get('name', ''))
        self.cmb_property.setCurrentText(data.get('property', ''))
        self.property_changed(0)
        self.txt_initial_value.setText(data.get('initial_value', ''))
        self.txt_expression.setText(data.get('expression', ''))

        channels = data.get('channels', [])
        self.tbl_channels.clearContents()
        self.tbl_channels.setRowCount(len(channels))
        vlabel = [str(i) for i in range(len(channels))]
        self.tbl_channels.setVerticalHeaderLabels(vlabel)
        ch_choices = {True: QtCore.Qt.Checked, False: QtCore.Qt.Unchecked}
        for row, ch in enumerate(channels):
            ch_name = ch.get('channel', '')
            ch_tr = ch.get('trigger', False)
            ch_use_enum = ch.get('use_enum', False)
            self.tbl_channels.setItem(row, 0,
                                      QtWidgets.QTableWidgetItem(str(ch_name)))
            checkbox_trigger = QtWidgets.QTableWidgetItem()
            checkbox_trigger.setCheckState(ch_choices[ch_tr])
            self.tbl_channels.setItem(row, 1, checkbox_trigger)
            checkbox_use_enum = QtWidgets.QTableWidgetItem()
            checkbox_use_enum.setCheckState(ch_choices[ch_use_enum])
            self.tbl_channels.setItem(row, 2, checkbox_use_enum)
        self.frm_edit.setEnabled(True)
        self.loading_data = False

    def add_rule(self):
        """Add a new rule to the list of rules."""
        default_name = "New Rule"
        data = {"name": default_name,
                "property": self.default_property,
                "initial_value": "",
                "expression": "",
                "channels": []
                }
        self.rules.append(data)
        self.lst_rule_item = QtWidgets.QListWidgetItem()
        self.lst_rule_item.setText(default_name)
        self.lst_rules.addItem(self.lst_rule_item)
        self.lst_rules.setCurrentItem(self.lst_rule_item)
        self.load_from_list()
        self.txt_name.setFocus()

    def get_current_index(self):
        """
        Calculate and return the selected index from the list of rules.

        Returns
        -------
        int
            The index selected at the list of rules or -1 in case the item
            does not exist.
        """
        if self.lst_rule_item is None:
            return -1
        return self.lst_rules.indexFromItem(self.lst_rule_item).row()

    def change_entry(self, entry, value):
        """
        Change an entry at the rules dictionary.

        Parameters
        ----------
        entry : str
            The key for the dictionary
        value : any
            The value to set on the key. It can be any type, depending on the
            key.

        Returns
        -------
        None
        """
        idx = self.get_current_index()
        self.rules[idx][entry] = value

    def del_rule(self):
        """Delete the rule selected in the rules list."""
        idx = self.get_current_index()
        if idx < 0:
            return
        name = self.lst_rule_item.text()

        confirm_message = "Are you sure you want to delete Rule: {}?".format(
            name)
        reply = QtWidgets.QMessageBox().question(self, 'Message',
                                             confirm_message,
                                             QtWidgets.QMessageBox.Yes,
                                             QtWidgets.QMessageBox.No)

        if reply == QtWidgets.QMessageBox.Yes:
            self.lst_rules.takeItem(idx)
            self.lst_rules.clearSelection()
            self.rules.pop(idx)
            self.clear_form()

    def add_channel(self):
        """Add a new empty channel to the table."""
        self.loading_data = True

        # Make the first entry be checked as suggestion
        if self.tbl_channels.rowCount() == 0:
            state = QtCore.Qt.Checked
        else:
            state = QtCore.Qt.Unchecked

        self.tbl_channels.insertRow(self.tbl_channels.rowCount())
        row = self.tbl_channels.rowCount() - 1
        self.tbl_channels.setItem(row, 0, QtWidgets.QTableWidgetItem(""))
        checkbox_trigger = QtWidgets.QTableWidgetItem()
        checkbox_trigger.setCheckState(state)
        checkbox_trigger.setFlags(QtCore.Qt.ItemIsEnabled|QtCore.Qt.ItemIsSelectable|QtCore.Qt.ItemIsUserCheckable)
        self.tbl_channels.setItem(row, 1, checkbox_trigger)
        checkbox_use_enum = QtWidgets.QTableWidgetItem()
        checkbox_use_enum.setCheckState(QtCore.Qt.Checked)
        checkbox_use_enum.setFlags(QtCore.Qt.ItemIsEnabled|QtCore.Qt.ItemIsSelectable|QtCore.Qt.ItemIsUserCheckable)
        self.tbl_channels.setItem(row, 2, checkbox_use_enum)
        vlabel = [str(i) for i in range(self.tbl_channels.rowCount())]
        self.tbl_channels.setVerticalHeaderLabels(vlabel)
        self.loading_data = False
        self.tbl_channels_changed()

    def del_channel(self):
        """Delete the selected channel at the table."""
        items = self.tbl_channels.selectedIndexes()
        if len(items) == 0:
            return

        c = "channel" if len(items) == 1 else "channels"
        confirm_message = "Delete the selected {}?".format(c)
        reply = QtWidgets.QMessageBox().question(self, 'Message',
                                             confirm_message,
                                             QtWidgets.QMessageBox.Yes,
                                             QtWidgets.QMessageBox.No)

        if reply == QtWidgets.QMessageBox.Yes:
            for itm in reversed(items):
                row = itm.row()
                self.tbl_channels.removeRow(row)

        self.tbl_channels_changed()

    def open_help(self, open=True):
        """
        Open the Help context for Rules.
        The documentation website prefix is given by the `PYDM_DOCS_URL`
        environmnet variable. If not defined it defaults to
        `https://slaclab.github.io/pydm`

        Parameters
        ----------
        open : bool
            Whether or not we should use the web browser to open the page.
        """
        docs_url = os.getenv("PYDM_DOCS_URL", None)
        if docs_url is None:
            docs_url = "https://slaclab.github.io/pydm"
        expression_url = "widgets/widget_rules/index.html"
        help_url = "{}/{}".format(docs_url, expression_url)
        if open:
            webbrowser.open(help_url, new=2, autoraise=True)
        else:
            return help_url

    def name_changed(self):
        """Callback executed when the rule name is changed."""
        if not self.lst_rule_item:
            return
        self.lst_rule_item.setText(self.txt_name.text())
        self.change_entry("name", self.txt_name.text())

    def property_changed(self, index):
        """Callback executed when the property is selected."""
        try:
            prop = self.cmb_property.currentData()
            self.lbl_expected_type.setText(prop[1].__name__)
            idx = self.get_current_index()
            self.change_entry("property", self.cmb_property.currentText())
        except:
            self.lbl_expected_type.setText("")

    def tbl_channels_changed(self, topleft=None, bottomright=None, roles=None):
        """Callback executed when the channels in the table are modified."""
        if self.loading_data:
            return

        new_channels = []

        for row in range(self.tbl_channels.rowCount()):
            ch = self.tbl_channels.item(row, 0).text()
            tr = self.tbl_channels.item(row,
                                        1).checkState() == QtCore.Qt.Checked
            en = self.tbl_channels.item(row,
                                        2).checkState() == QtCore.Qt.Checked
            new_channels.append({"channel": ch, "trigger": tr, "use_enum": en})

        self.change_entry("channels", new_channels)

    def expression_changed(self):
        """Callback executed when the expression is modified."""
        self.change_entry("expression", self.txt_expression.text())

    def initial_value_changed(self):
        """Callback executed when the initial value is modified"""
        self.change_entry("initial_value", self.txt_initial_value.text())

    @staticmethod
    def is_data_valid(rules):
        """
        Sanity check the form data.

        Parameters
        ----------
        rules : dict
            Dictionary with the rule defintions.

        Returns
        -------
        tuple : (bool, str)
            True and "" in case there is no error or False and the error message
            otherwise.
        """
        errors = []
        for idx, rule in enumerate(rules):
            name = rule.get("name")
            expression = rule.get("expression")
            channels = rule.get("channels", [])

            if name is None or name == "":
                errors.append("Rule #{} has no name.".format(idx + 1))
            if expression is None or expression == "":
                errors.append("Rule #{} has no expression.".format(idx + 1))
            if len(channels) == 0:
                errors.append("Rule #{} has no channel.".format(idx + 1))
            else:
                found_trigger = False
                for ch_idx, ch in enumerate(channels):
                    if not ch.get("channel", ""):
                        errors.append(
                            "Rule #{} - Ch. #{} has no channel.".format(idx + 1,
                                                                        ch_idx))
                    if ch.get("trigger", False) and not found_trigger:
                        found_trigger = True

                if not found_trigger:
                    errors.append(
                        "Rule #{} has no channel for trigger.".format(idx + 1))

        if len(errors) > 0:
            error_msg = os.linesep.join(errors)
            return False, error_msg

        return True, ""

    @QtCore.Slot()
    def saveChanges(self):
        """Save the new rules at the widget `rules` property."""
        # If the form is being edited, we make sure self.rules has all the
        # latest values from the form before we try to validate.  This fixes
        # a problem where the last form item change wouldn't get saved unless
        # the user knew to hit 'enter' or leave the field to end editing before
        # hitting save.
        if self.frm_edit.isEnabled():
            self.expression_changed()
            self.name_changed()
            self.tbl_channels_changed()
        status, message = RulesEditor.is_data_valid(self.rules)
        if status:
            data = json.dumps(self.rules)
            formWindow = QtDesigner.QDesignerFormWindowInterface.findFormWindow(self.widget)
            if formWindow:
                formWindow.cursor().setProperty("rules", data)
            self.accept()
        else:
            QtWidgets.QMessageBox.critical(self, "Error Saving", message,
                                       QtWidgets.QMessageBox.Ok)

    @QtCore.Slot()
    def cancelChanges(self):
        """Abort the changes and close the dialog."""
        self.close()



from qtpy.QtWidgets import QDoubleSpinBox, QApplication, QLineEdit
from qtpy.QtCore import Property, QEvent, Qt
from .base import PyDMWritableWidget, TextFormatter


class PyDMSpinbox(QDoubleSpinBox, TextFormatter, PyDMWritableWidget):
    """
    A QDoubleSpinBox with support for Channels and more from PyDM.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    def __init__(self, parent=None, init_channel=None):
        QDoubleSpinBox.__init__(self, parent)
        PyDMWritableWidget.__init__(self, init_channel=init_channel)
        self.valueBeingSet = False
        self.setEnabled(False)
        self._alarm_sensitive_border = False
        self._show_step_exponent = True
        self._write_on_press = False
        self.step_exponent = 0
        self.setDecimals(0)
        self.app = QApplication.instance()
        self.setAccelerated(True)

        # install an event filter on the QDoubleSpinBox's children
        # in order to catch the click events
        child = self.findChild(QLineEdit)
        child.installEventFilter(self)


    def stepBy(self, step):
        """
        Method triggered whenever user triggers a step. If the writeOnPress property
        is enabled, the updated value will be sent.

        Parameters
        ----------
        step: int
        """
        super(PyDMSpinbox, self).stepBy(step)
        if self._write_on_press:
            self.send_value()


    def keyPressEvent(self, ev):
        """
        Method invoked when a key press event happens on the QDoubleSpinBox.

        For PyDMSpinBox we are interested on the Keypress events for:
            - CTRL + Left/Right : Increase or Decrease the step exponent;
            - Up / Down : Add or Remove `singleStep` units to the value;
            - PageUp / PageDown : Add or Remove 10 times `singleStep` units
              to the value;
            - Return or Enter : Send the value to the channel using the
              `send_value_signal`.

        Parameters
        ----------
        ev : QEvent
        """
        ctrl_hold = self.app.queryKeyboardModifiers() == Qt.ControlModifier
        if ctrl_hold and (ev.key() in (Qt.Key_Left, Qt.Key_Right)):
            self.step_exponent += 1 if ev.key() == Qt.Key_Left else -1
            self.step_exponent = max(-self.decimals(), self.step_exponent)
            self.update_step_size()

        elif ev.key() in (Qt.Key_Return, Qt.Key_Enter):
            self.send_value()
    
        else:
            super(PyDMSpinbox, self).keyPressEvent(ev)
            

    def widget_ctx_menu(self):
        """
        Fetch the Widget specific context menu which will be populated with additional tools by `assemble_tools_menu`.

        Returns
        -------
        QMenu or None
            If the return of this method is None a new QMenu will be created by `assemble_tools_menu`.
        """
        def toggle_step():
            self.showStepExponent = not self.showStepExponent

        def toggle_write():
            self.writeOnPress = not self.writeOnPress

        menu = self.lineEdit().createStandardContextMenu()
        menu.addSeparator()
        ac = menu.addAction('Toggle Show Step Size')
        ac.triggered.connect(toggle_step)

        ac_write = menu.addAction('Toggle Write On Press')
        ac_write.triggered.connect(toggle_write)

        return menu

    def update_step_size(self):
        """
        Update the Single Step size on the QDoubleSpinBox.
        """
        self.setSingleStep(10 ** self.step_exponent)
        self.update_format_string()

    def update_format_string(self):
        """
        Reconstruct the format string to be used when representing the
        output value.

        Returns
        -------
        None
        """
        if self._show_units:
            units = " {}".format(self._unit)
        else:
            units = ""

        if self._show_step_exponent:
            self.setSuffix("{0} Step: 1E{1}".format(units, self.step_exponent))
            self.lineEdit().setToolTip("")
        else:
            self.setSuffix(units)
            self.lineEdit().setToolTip('Step: 1E{0:+d}'.format(self.step_exponent))

    def value_changed(self, new_val):
        """
        Callback invoked when the Channel value is changed.

        Parameters
        ----------
        new_val : int or float
            The new value from the channel.
        """
        if new_val is None:
            # SpinBox is unable to work with None and
            # but sometimes it can arrive as an initial value
            return
        super(PyDMSpinbox, self).value_changed(new_val)
        self.valueBeingSet = True
        self.setValue(new_val)
        self.valueBeingSet = False

    def send_value(self):
        """
        Method invoked to send the current value on the QDoubleSpinBox to
        the channel using the `send_value_signal`.
        """
        value = QDoubleSpinBox.value(self)
        if not self.valueBeingSet:
            self.send_value_signal[float].emit(value)

    def ctrl_limit_changed(self, which, new_limit):
        """
        Callback invoked when the Channel receives new control limit
        values.

        Parameters
        ----------
        which : str
            Which control limit was changed. "UPPER" or "LOWER"
        new_limit : float
            New value for the control limit
        """
        super(PyDMSpinbox, self).ctrl_limit_changed(which, new_limit)
        if which == "UPPER":
            self.setMaximum(new_limit)
        else:
            self.setMinimum(new_limit)

    def precision_changed(self, new_precision):
        """
        Callback invoked when the Channel has new precision value.
        This callback also triggers an update_format_string call so the
        new precision value is considered.

        Parameters
        ----------
        new_precison : int or float
            The new precision value
        """
        super(PyDMSpinbox, self).precision_changed(new_precision)
        self.setDecimals(self.precision)

    @Property(int)
    def precision(self):
        if self.precisionFromPV:
            return self._prec
        else:
            return self._user_prec

    @precision.setter
    def precision(self, new_prec):
        if self.precisionFromPV:
            return
        if new_prec and self._user_prec != int(new_prec) and new_prec >= 0:
            self._user_prec = int(new_prec)
            self.value_changed(self.value)
            self.setDecimals(new_prec)

    @Property(bool)
    def showStepExponent(self):
        """
        Whether to show or not the step exponent

        Returns
        -------
        bool
        """
        return self._show_step_exponent

    @showStepExponent.setter
    def showStepExponent(self, val):
        """
        Whether to show or not the step exponent

        Parameters
        ----------
        val : bool
        """
        self._show_step_exponent = val
        self.update_format_string()

    @Property(bool)
    def writeOnPress(self):
        """
        Whether to write value on key press

        Returns
        -------
        bool
        """
        return self._write_on_press

    @writeOnPress.setter
    def writeOnPress(self, val):
        """
        Whether value to write on key press.

        Parameters
        ----------
        val : bool
        """
        self._write_on_press = val



# Unit Tests for the Shell Command widget class

import os
import pytest

import platform
from logging import ERROR

from qtpy import QtCore
from qtpy.QtCore import QSize
from qtpy.QtWidgets import QMenu, QAction

from ...widgets.shell_command import PyDMShellCommand
from ...utilities import IconFont


# --------------------
# POSITIVE TEST CASES
# --------------------

@pytest.mark.parametrize("command, title", [
    ("foo", None),
    ("", None),
    (None, None),
    (["foo", "bar"], ["A", "B"]),
])
def test_construct(qtbot, command, title):
    """
    Test the construct of the widget.

    Expectations:
    The widget is initialized, and the commands and titles match the
    constructor arguments provided.  Also tests the icon and cursor pixmaps.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    command : str
        The shell command(s) to be executed when the widget is pressed
    title : str
        The titles for the shell commands.  Really only relevant if
        len(commands) > 1

    """
    pydm_shell_command = PyDMShellCommand(command=command, title=title)
    qtbot.addWidget(pydm_shell_command)
    if command and isinstance(command, str):
        assert pydm_shell_command._commands == [command]
    elif command:
        assert pydm_shell_command._commands == command
    else:
        assert pydm_shell_command._commands == []
    if title and isinstance(title, str):
        assert pydm_shell_command._titles == [title]
    elif title:
        assert pydm_shell_command._titles == title
    else:
        assert pydm_shell_command._titles == []
    
    DEFAULT_ICON_NAME = "cog"
    DEFAULT_ICON_SIZE = QSize(16, 16)

    default_icon = IconFont().icon(DEFAULT_ICON_NAME)

    default_icon_pixmap = default_icon.pixmap(DEFAULT_ICON_SIZE)
    shell_cmd_icon_pixmap = pydm_shell_command.icon().pixmap(DEFAULT_ICON_SIZE)

    assert shell_cmd_icon_pixmap.toImage() == default_icon_pixmap.toImage()
    assert pydm_shell_command.cursor().pixmap().toImage() == default_icon_pixmap.toImage()

def test_deprecated_command_property_with_no_commands(qtbot):
    pydm_shell_command = PyDMShellCommand()
    qtbot.addWidget(pydm_shell_command)
    pydm_shell_command.command = "test"
    assert pydm_shell_command.commands == ["test"]

def test_deprecated_command_property_with_commands(qtbot):
    pydm_shell_command = PyDMShellCommand()
    qtbot.addWidget(pydm_shell_command)
    existing_commands = ["existing", "commands"]
    pydm_shell_command.commands = existing_commands
    pydm_shell_command.command = "This shouldn't work"
    assert pydm_shell_command.commands == existing_commands

def test_no_crash_without_any_commands(qtbot):
     pydm_shell_command = PyDMShellCommand()
     qtbot.addWidget(pydm_shell_command)
     pydm_shell_command.commands = None
     qtbot.mouseClick(pydm_shell_command, QtCore.Qt.LeftButton)

def test_no_crash_with_none_command(qtbot):
     pydm_shell_command = PyDMShellCommand()
     qtbot.addWidget(pydm_shell_command)
     pydm_shell_command.command = None
     qtbot.mouseClick(pydm_shell_command, QtCore.Qt.LeftButton)

@pytest.mark.parametrize("cmd, retcode, stdout", [
    (["choice /c yn /d n /t 0"] if platform.system() == "Windows" else ["sleep 0"],
        [2] if platform.system() == "Windows" else [0], ["[Y,N]?N"] if platform.system() == "Windows" else [""]),
    (["pydm_shell_invalid_command_test invalid command"], [None], [""]),
    (["echo hello", "echo world"], [0, 0], ["hello", "world"])
])
def test_mouse_release_event(qtbot, caplog, cmd, retcode, stdout):
    """
    Test to ensure the widget's triggering of the Mouse Release event.
    
    Expectations if len(cmd) == 1:
    1. The mouse release will trigger the current shell command being assigned to the widget to execute
    2. If the command is not valid, there will be an error message in the log
    3. If the command is valid, there will be output in stdout (the result could be a success or failure, but at least
        the command will send out text to stdout)
    4. If the command is None or empty, there will be no output to stdout.
    
    Expectations if len(cmd) > 1:
    1. The mouse press will cause the widget to set its 'menu' attribute to an instance of QMenu.
    2. Triggering each menu item runs the right command.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    caplog : fixture
        To capture the log messages
    cmd : str
        The shell command for the widget to execute
    retcode : int
        The expected exit code.
    stdout : str
        The expected stdout for the command.
    """
    pydm_shell_command = PyDMShellCommand()
    qtbot.addWidget(pydm_shell_command)

    def check_command_output(command, expected_retcode, expected_stdout):
        if "invalid" not in command:
            stdout, stderr = pydm_shell_command.process.communicate()
            assert pydm_shell_command.process.returncode == expected_retcode
            assert expected_stdout in str(stdout)
        else:
            for record in caplog.records:
                assert record.levelno == ERROR
            assert "Error in shell command" in caplog.text

    pydm_shell_command.commands = cmd
    
    if len(cmd) > 1:
        for current_command, expected_retcode, expected_stdout in zip(cmd, retcode, stdout):
            # We can't actually do the click and show the menu - it halts the test and waits
            # for user input.  So instead, we'll force trigger _rebuild_menu().
            pydm_shell_command._rebuild_menu()
            assert isinstance(pydm_shell_command.menu(), QMenu)
            actions = pydm_shell_command.menu().findChildren(QAction)
            assert current_command in [a.text() for a in actions]
            action_for_current_command = [a for a in actions if a.text() == current_command][0]
            action_for_current_command.trigger()
            check_command_output(current_command, expected_retcode, expected_stdout)
    elif len(cmd) == 1:
        qtbot.mouseClick(pydm_shell_command, QtCore.Qt.LeftButton)
        check_command_output(cmd[0], retcode[0], stdout[0])
    else:
        qtbot.mouseClick(pydm_shell_command, QtCore.Qt.LeftButton)
        assert pydm_shell_command.process is None


@pytest.mark.parametrize("allow_multiple", [
    True,
    False,
])
def test_execute_multiple_commands(qtbot, signals, caplog, allow_multiple):
    """
    Test the widget's ability to execute multiple shell commands when this setting is enabled.

    Expectations:
    1. If the multiple execution setting is enabled, the widget must execute another shell command while the previous
        one is still running.
    2. If the multiple execution setting is disabled, the widget will log the error "Command already active."

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    caplog : fixture
        To capture the log messages
    allow_multiple : bool
        True if multiple command executions are allowed for the widget; False otherwise
    """
    pydm_shell_command = PyDMShellCommand()
    qtbot.addWidget(pydm_shell_command)

    pydm_shell_command._allow_multiple = allow_multiple

    cmd = "choice /c yn /d y /t 1" if platform.system() == "Windows" else "sleep 0.1"
    pydm_shell_command.execute_command(cmd)
    pydm_shell_command.execute_command(cmd)

    if not allow_multiple:
        for record in caplog.records:
            assert record.levelno == ERROR
        assert "already active" in caplog.text
    else:
        assert not caplog.text


def test_env_var(qtbot):
    """
    Test to ensure the widget can handle commands with environment variables
    in it.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    pydm_shell_command = PyDMShellCommand()
    qtbot.addWidget(pydm_shell_command)

    cmd = "echo Test: $PATH"
    if platform.system() == 'Windows':
        cmd = "echo Test: %PATH%"

    pydm_shell_command.commands = [cmd]
    qtbot.mouseClick(pydm_shell_command, QtCore.Qt.LeftButton)
    stdout, stderr = pydm_shell_command.process.communicate()
    assert pydm_shell_command.process.returncode == 0
    if platform.system() == 'Windows':
        # Windows changes C:\\ to C:\\\\
        assert os.getenv("PATH") in str(stdout).replace('\\\\', '\\')
    else:
        assert "Test: {}".format(os.getenv("PATH")) in str(stdout)



import sys
import argparse
import logging


def main():
    logger = logging.getLogger('')
    handler = logging.StreamHandler()
    formatter = logging.Formatter('[%(asctime)s] [%(levelname)-8s] - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel("INFO")
    handler.setLevel("INFO")

    from pydm.utilities import setup_renderer

    setup_renderer()

    try:
        """
        We must import QtWebEngineWidgets before creating a QApplication
        otherwise we get the following error if someone adds a WebView at Designer:
        ImportError: QtWebEngineWidgets must be imported before a QCoreApplication instance is created
        """
        from qtpy import QtWebEngineWidgets
    except ImportError:
        logger.debug('QtWebEngine is not supported.')

    import pydm
    from pydm.utilities.macro import parse_macro_string

    parser = argparse.ArgumentParser(description="Python Display Manager")
    parser.add_argument(
        'displayfile',
        help='A PyDM file to display.' +
             '    Can be either a Qt .ui file, or a Python file.',
        nargs='?',
        default=None
        )
    parser.add_argument(
        '--perfmon',
        action='store_true',
        help='Enable performance monitoring,' +
             ' and print CPU usage to the terminal.'
        )
    parser.add_argument(
        '--hide-nav-bar',
        action='store_true',
        help='Start PyDM with the navigation bar hidden.'
        )
    parser.add_argument(
        '--hide-menu-bar',
        action='store_true',
        help='Start PyDM with the menu bar hidden.'
        )
    parser.add_argument(
        '--hide-status-bar',
        action='store_true',
        help='Start PyDM with the status bar hidden.'
        )
    parser.add_argument(
        '--fullscreen',
        action='store_true',
        help='Start PyDM in full screen mode.'
        )
    parser.add_argument(
        '--read-only',
        action='store_true',
        help='Start PyDM in a Read-Only mode.'
        )
    parser.add_argument(
        '--log_level',
        help='Configure level of log display',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        default='INFO'
        )
    parser.add_argument('--version', action='version',
                    version='PyDM {version}'.format(version=pydm.__version__),
                    help="Show PyDM's version number and exit.")
    parser.add_argument(
        '-m', '--macro',
        help='Specify macro replacements to use, in JSON object format.' +
             '    Reminder: JSON requires double quotes for strings, ' +
             'so you should wrap this whole argument in single quotes.' +
             '  Example: -m \'{"sector": "LI25", "facility": "LCLS"}\'' +
             '--or-- specify macro replacements as KEY=value pairs ' +
             ' using a comma as delimiter  If you want to uses spaces ' +
             ' after the delimiters or around the = signs, ' +
             ' wrap the entire set with quotes ' +
             '  Example: -m "sector = LI25, facility=LCLS"'
        )
    parser.add_argument(
        '--stylesheet',
        help='Specify the full path to a CSS stylesheet file, which' +
             ' can be used to customize the appearance of PyDM and' +
             ' Qt widgets.',
        default=None
        )
    parser.add_argument(
        'display_args',
        help='Arguments to be passed to the PyDM client application' +
             ' (which is a QApplication subclass).',
        nargs=argparse.REMAINDER,
        default=None
        )

    pydm_args = parser.parse_args()
    macros = None
    if pydm_args.macro is not None:
        macros = parse_macro_string(pydm_args.macro)

    if pydm_args.log_level:
        logger.setLevel(pydm_args.log_level)
        handler.setLevel(pydm_args.log_level)

    app = pydm.PyDMApplication(
        ui_file=pydm_args.displayfile,
        command_line_args=pydm_args.display_args,
        perfmon=pydm_args.perfmon,
        hide_nav_bar=pydm_args.hide_nav_bar,
        hide_menu_bar=pydm_args.hide_menu_bar,
        hide_status_bar=pydm_args.hide_status_bar,
        fullscreen=pydm_args.fullscreen,
        read_only=pydm_args.read_only,
        macros=macros,
        stylesheet_path=pydm_args.stylesheet
        )

    pydm.utilities.shortcuts.install_connection_inspector(
        parent=app.main_window)

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



from qtpy.QtCore import QModelIndex, QVariant
from .baseplot_table_model import BasePlotCurvesModel
from .baseplot_curve_editor import BasePlotCurveEditorDialog, PlotStyleColumnDelegate, RedrawModeColumnDelegate


class PyDMScatterPlotCurvesModel(BasePlotCurvesModel):
    """ This is the data model used by the waveform plot curve editor.
    It basically acts as a go-between for the curves in a plot, and
    QTableView items.
    """

    def __init__(self, plot, parent=None):
        super(PyDMScatterPlotCurvesModel, self).__init__(plot, parent=parent)
        self._column_names = ('Y Channel', 'X Channel') + self._column_names
        self._column_names += ('Redraw Mode', 'Buffer Size')

    def get_data(self, column_name, curve):
        if column_name == "Y Channel":
            if curve.y_address is None:
                return QVariant()
            return str(curve.y_address)
        elif column_name == "X Channel":
            if curve.x_address is None:
                return QVariant()
            return str(curve.x_address)
        elif column_name == "Redraw Mode":
            return curve.redraw_mode
        elif column_name == "Buffer Size":
            return curve.getBufferSize()
        return super(PyDMScatterPlotCurvesModel, self).get_data(
            column_name, curve)

    def set_data(self, column_name, curve, value):
        if column_name == "Y Channel":
            curve.y_address = str(value)
        elif column_name == "X Channel":
            curve.x_address = str(value)
        elif column_name == "Redraw Mode":
            curve.redraw_mode = int(value)
        elif column_name == "Buffer Size":
            curve.setBufferSize(int(value))
        else:
            return super(PyDMScatterPlotCurvesModel, self).set_data(
                column_name=column_name, curve=curve, value=value)
        return True

    def append(self, y_address=None, x_address=None, name=None, color=None):
        self.beginInsertRows(QModelIndex(), len(self._plot._curves),
                             len(self._plot._curves))
        self._plot.addChannel(y_address, x_address, name, color)
        self.endInsertRows()

    def removeAtIndex(self, index):
        self.beginRemoveRows(QModelIndex(), index.row(), index.row())
        self._plot.removeChannelAtIndex(index.row())
        self.endRemoveRows()


class ScatterPlotCurveEditorDialog(BasePlotCurveEditorDialog):
    """ScatterPlotCurveEditorDialog is a QDialog that is used in Qt Designer
    to edit the properties of the curves in a waveform plot.  This dialog is
    shown when you double-click the plot, or when you right click it and
    choose 'edit curves'.

    This thing is mostly just a wrapper for a table view, with a couple
    buttons to add and remove curves, and a button to save the changes."""
    TABLE_MODEL_CLASS = PyDMScatterPlotCurvesModel

    def __init__(self, plot, parent=None):
        super(ScatterPlotCurveEditorDialog, self).__init__(plot, parent)

        redraw_mode_delegate = RedrawModeColumnDelegate(self)
        self.table_view.setItemDelegateForColumn(self.table_model.getColumnIndex("Redraw Mode"), redraw_mode_delegate)

        plot_style_delegate = PlotStyleColumnDelegate(self, self.table_model, self.table_view)
        plot_style_delegate.hideColumns(hide_line_columns=False, hide_bar_columns=True)



import io
import sys
import traceback
from six.moves import queue
from collections import namedtuple

from qtpy import QtWidgets, QtCore

from .utilities import only_main_thread

"""
Utility functions which installs an exception hook and displays any global 
uncaught exception to operators.

excepthook is based on https://fman.io/blog/pyqt-excepthook/
raise_to_operator is based on https://github.com/pcdshub/typhos/blob/0837405e8d8b73227ac34eb69d2390eed85f3c65/typhos/utils.py#L318
"""


class ExceptionDispatcher(QtCore.QThread):
    """
    Singleton QTread class that receives and dispatch uncaught exceptions via
    the `newException` signal.
    """
    newException = QtCore.Signal(object)

    __instance = None

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = QtCore.QThread.__new__(ExceptionDispatcher,
                                                    *args, **kwargs)
            cls.__instance.__initialized = False
        return cls.__instance

    def __init__(self, *args, **kwargs):
        if self.__initialized:
            return
        super(ExceptionDispatcher, self).__init__(*args, **kwargs)
        self.__initialized = True
        self.app = QtWidgets.QApplication.instance()
        self.app.aboutToQuit.connect(self.requestInterruption)
        self._queue = queue.Queue()

    def add(self, exc_type, exc_value, exc_tb):
        """
        Add an uncaught exception into the Queue.

        Parameters
        ----------
        exc_type : type
            The exception type
        exc_value : Exception
            The exception object
        exc_tb : traceback
            The traceback namedtuple.
        """
        self._queue.put((exc_type, exc_value, exc_tb))

    def run(self):
        def create_entry(entry):
            # Use the Type and message to deduplicate the errors
            key = (entry[0], str(entry[1]))
            return {key: entry}

        while not self.isInterruptionRequested():
            bucket = {}
            # Wait until we have an error
            data = self._queue.get(block=True, timeout=None)  # Block forever
            # Wait a bit to check if we have more
            self.msleep(200)

            bucket.update(create_entry(data))

            # Collect all errors
            while not self._queue.empty():
                data = self._queue.get(block=False)
                bucket.update(create_entry(data))

            # Emit unique errors to listeners
            for _, v in bucket.items():
                self.newException.emit(data)


class DefaultExceptionNotifier(QtCore.QObject):
    """
    Default handler for exceptions which prints at the shell the exception
    and also calls `raise_to_operator` to display a dialog with the exception
    information.
    """
    __instance = None

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = QtCore.QObject.__new__(DefaultExceptionNotifier,
                                                    *args, **kwargs)
            cls.__instance.__initialized = False
        return cls.__instance

    def __init__(self, *args, **kwargs):
        if self.__initialized:
            return
        super(DefaultExceptionNotifier, self).__init__(*args, **kwargs)
        self.__initialized = True
        ExceptionDispatcher().newException.connect(self.receiveException)

    @QtCore.Slot(tuple)
    def receiveException(self, exception_data):
        exc_type, exc_value, enriched_tb = exception_data
        traceback.print_exception(exc_type, exc_value, enriched_tb)
        raise_to_operator(exc_value)


_old_excepthook = None

fake_tb = namedtuple(
    'fake_tb', ('tb_frame', 'tb_lasti', 'tb_lineno', 'tb_next')
)


def excepthook(exc_type, exc_value, exc_tb):
    enriched_tb = _add_missing_frames(exc_tb) if exc_tb else exc_tb
    ExceptionDispatcher().add(exc_type, exc_value, enriched_tb)


def _add_missing_frames(tb):
    """
    Originally from: https://fman.io/blog/pyqt-excepthook/

    When an exception occurs in Python, sys.excepthook(...) is called with an
    exc_tb parameter. This parameter contains the information for each of the
    lines in the Tracebacks shown above. The reason why the first version of
    our code did not include f() in the traceback was that it did not appear
    in exc_tb.

    To fix the problem, our additional excepthook code above creates a "fake"
    traceback that includes the missing entries. Fortunately, the necessary
    information is available in the .tb_frame property of the original
    traceback. Finally, the default sys.__excepthook__(...) does not work with
    fake data, so we need to call traceback.print_exception(...) instead.
    """
    result = fake_tb(tb.tb_frame, tb.tb_lasti,
                     tb.tb_lineno, tb.tb_next)
    frame = tb.tb_frame.f_back
    while frame:
        result = fake_tb(frame, frame.f_lasti,
                         frame.f_lineno, result)
        frame = frame.f_back
    return result


def install(use_default_handler=True):
    """
    Install the exception hook handler.
    If use_dialog is specified, a QMessageBox will also be presented.

    Parameters
    ----------
    use_default_handler : bool
        Whether or not to use the default handler. If not using it, users
        must connect to the `ExceptionDispatcher.newException` signal.
    """
    global _old_excepthook

    if _old_excepthook is None:
        _old_excepthook = sys.excepthook

    dispatcher = ExceptionDispatcher()
    if dispatcher.isRunning():
        return
    if use_default_handler:
        handler = DefaultExceptionNotifier()
    dispatcher.start()
    sys.excepthook = excepthook


def uninstall():
    """
    Uninstall the exception hook handler and revert to the previous value.
    """
    global _old_excepthook

    if _old_excepthook is None:
        return
    sys.excepthook = _old_excepthook
    ExceptionDispatcher().requestInterruption()
    _old_excepthook = None


@only_main_thread
def raise_to_operator(exc):
    """Utility function to show an Exception to the operator"""
    err_msg = QtWidgets.QMessageBox()
    err_msg.setText('{}: {}'.format(exc.__class__.__name__, exc))
    err_msg.setWindowTitle(type(exc).__name__)
    err_msg.setIcon(QtWidgets.QMessageBox.Critical)
    handle = io.StringIO()
    traceback.print_tb(exc.__traceback__, file=handle)
    handle.seek(0)
    err_msg.setDetailedText(handle.read())
    err_msg.exec_()
    return err_msg



import pytest

from ...widgets.byte import PyDMByteIndicator


@pytest.mark.parametrize("shift, value, expected", [
    (0, 0, (False, False, False)),
    (0, 5, (True, False, True)),
    (0, -5, (False, False, False)),

    (1, 0, (False, False, False)),
    (1, 4, (False, True, False)),
    (1, -5, (False, False, False)),

    (-1, 0, (False, False, False)),
    (-1, 1, (False, True, False)),
    (-1, -5, (False, False, False)),
])
def test_value_shift(qtbot, signals, shift, value, expected):
    """
    Test the widget's handling of the value changed event affected by predefined shift.

    Expectations:
    1. Value coming from the control system is correctly shifted by the predefined value
    2. Resulting value should be non-negative

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    value : int
        The value to be displayed by the widget
    shift : int
        The value's bit shift
    expected : int
        Expected resulting value
    """
    num_bits = len(expected)
    pydm_byte = PyDMByteIndicator()
    qtbot.addWidget(pydm_byte)
    pydm_byte.numBits = num_bits
    pydm_byte.shift = shift
    pydm_byte._connected = True

    signals.new_value_signal[type(value)].connect(pydm_byte.channelValueChanged)
    signals.new_value_signal[type(value)].emit(value)

    for i, bit, indicator in zip(range(num_bits), expected, pydm_byte._indicators):
        expected_color = pydm_byte.onColor if bit else pydm_byte.offColor
        assert indicator._brush.color().name() == expected_color.name(), "Failed to match bit#{}".format(i)



import weakref
import threading

from numpy import ndarray

from ..utilities.remove_protocol import protocol_and_address
from qtpy.QtCore import Signal, QObject, Qt
from qtpy.QtWidgets import QApplication
from .. import config


class PyDMConnection(QObject):
    new_value_signal = Signal([float], [int], [str], [ndarray], [bool])
    connection_state_signal = Signal(bool)
    new_severity_signal = Signal(int)
    write_access_signal = Signal(bool)
    enum_strings_signal = Signal(tuple)
    unit_signal = Signal(str)
    prec_signal = Signal(int)
    upper_ctrl_limit_signal = Signal([float], [int])
    lower_ctrl_limit_signal = Signal([float], [int])

    def __init__(self, channel, address, protocol=None, parent=None):
        super(PyDMConnection, self).__init__(parent)
        self.protocol = protocol
        self.address = address
        self.connected = False
        self.value = None
        self.listener_count = 0
        self.app = QApplication.instance()

    def add_listener(self, channel):
        self.listener_count = self.listener_count + 1
        if channel.connection_slot is not None:
            self.connection_state_signal.connect(channel.connection_slot, Qt.QueuedConnection)

        if channel.value_slot is not None:
            try:
                self.new_value_signal[int].connect(channel.value_slot, Qt.QueuedConnection)
            except TypeError:
                pass
            try:
                self.new_value_signal[float].connect(channel.value_slot, Qt.QueuedConnection)
            except TypeError:
                pass
            try:
                self.new_value_signal[str].connect(channel.value_slot, Qt.QueuedConnection)
            except TypeError:
                pass
            try:
                self.new_value_signal[ndarray].connect(channel.value_slot, Qt.QueuedConnection)
            except TypeError:
                pass
            try:
                self.new_value_signal[bool].connect(channel.value_slot, Qt.QueuedConnection)
            except TypeError:
                pass

        if channel.severity_slot is not None:
            self.new_severity_signal.connect(channel.severity_slot, Qt.QueuedConnection)

        if channel.write_access_slot is not None:
            self.write_access_signal.connect(channel.write_access_slot, Qt.QueuedConnection)

        if channel.enum_strings_slot is not None:
            self.enum_strings_signal.connect(channel.enum_strings_slot, Qt.QueuedConnection)

        if channel.unit_slot is not None:
            self.unit_signal.connect(channel.unit_slot, Qt.QueuedConnection)

        if channel.upper_ctrl_limit_slot is not None:
            self.upper_ctrl_limit_signal.connect(channel.upper_ctrl_limit_slot, Qt.QueuedConnection)

        if channel.lower_ctrl_limit_slot is not None:
            self.lower_ctrl_limit_signal.connect(channel.lower_ctrl_limit_slot, Qt.QueuedConnection)

        if channel.prec_slot is not None:
            self.prec_signal.connect(channel.prec_slot, Qt.QueuedConnection)

    def remove_listener(self, channel, destroying=False):
        if not destroying:
            if channel.connection_slot is not None:
                try:
                    self.connection_state_signal.disconnect(channel.connection_slot)
                except TypeError:
                    pass

            if channel.value_slot is not None:
                try:
                    self.new_value_signal[int].disconnect(channel.value_slot)
                except TypeError:
                    pass
                try:
                    self.new_value_signal[float].disconnect(channel.value_slot)
                except TypeError:
                    pass
                try:
                    self.new_value_signal[str].disconnect(channel.value_slot)
                except TypeError:
                    pass
                try:
                    self.new_value_signal[ndarray].disconnect(channel.value_slot)
                except TypeError:
                    pass
                try:
                    self.new_value_signal[bool].disconnect(channel.value_slot)
                except TypeError:
                    pass

            if channel.severity_slot is not None:
                try:
                    self.new_severity_signal.disconnect(channel.severity_slot)
                except (KeyError, TypeError):
                    pass

            if channel.write_access_slot is not None:
                try:
                    self.write_access_signal.disconnect(channel.write_access_slot)
                except (KeyError, TypeError):
                    pass

            if channel.enum_strings_slot is not None:
                try:
                    self.enum_strings_signal.disconnect(channel.enum_strings_slot)
                except (KeyError, TypeError):
                    pass

            if channel.unit_slot is not None:
                try:
                    self.unit_signal.disconnect(channel.unit_slot)
                except (KeyError, TypeError):
                    pass

            if channel.upper_ctrl_limit_slot is not None:
                try:
                    self.upper_ctrl_limit_signal.disconnect(channel.upper_ctrl_limit_slot)
                except (KeyError, TypeError):
                    pass

            if channel.lower_ctrl_limit_slot is not None:
                try:
                    self.lower_ctrl_limit_signal.disconnect(channel.lower_ctrl_limit_slot)
                except (KeyError, TypeError):
                    pass

            if channel.prec_slot is not None:
                try:
                    self.prec_signal.disconnect(channel.prec_slot)
                except (KeyError, TypeError):
                    pass

        self.listener_count = self.listener_count - 1
        if self.listener_count < 1:
            self.close()

    def close(self):
        pass

class PyDMPlugin(object):
    protocol = None
    connection_class = PyDMConnection
    designer_online_by_default = False

    def __init__(self):
        self.connections = {}
        self.channels = weakref.WeakSet()
        self.lock = threading.Lock()

    @staticmethod
    def get_address(channel):
        return protocol_and_address(channel.address)[1]

    @staticmethod
    def get_connection_id(channel):
        return PyDMPlugin.get_address(channel)

    def add_connection(self, channel):
        from pydm.utilities import is_qt_designer
        with self.lock:
            connection_id = self.get_connection_id(channel)
            address = self.get_address(channel)
            # If this channel is already connected to this plugin lets ignore
            if channel in self.channels:
                return

            if (is_qt_designer() and not config.DESIGNER_ONLINE and
                    not self.designer_online_by_default):
                return

            self.channels.add(channel)
            if connection_id in self.connections:
                self.connections[connection_id].add_listener(channel)
            else:
                self.connections[connection_id] = self.connection_class(
                    channel, address, self.protocol)

    def remove_connection(self, channel, destroying=False):
        with self.lock:
            connection_id = self.get_connection_id(channel)
            if connection_id in self.connections and channel in self.channels:
                self.connections[connection_id].remove_listener(
                    channel, destroying=destroying)
                self.channels.remove(channel)
                if self.connections[connection_id].listener_count < 1:
                    self.connections[connection_id].deleteLater()
                    del self.connections[connection_id]




from ...utilities.remove_protocol import remove_protocol


def test_remove_protocol():
    out = remove_protocol('foo://bar')
    assert (out == 'bar')

    out = remove_protocol('bar')
    assert (out == 'bar')

    out = remove_protocol('foo://bar://foo2')
    assert (out == 'bar://foo2')



from ..widgets.qtplugin_base import qtplugin_factory

def test_import_byte_plugin():
    # Byte plugin
    from ..widgets.byte import PyDMByteIndicator
    PyDMByteIndicatorPlugin = qtplugin_factory(PyDMByteIndicator)

def test_import_checkbox_plugin():
    # Checkbox plugin
    from ..widgets.checkbox import PyDMCheckbox
    PyDMCheckboxPlugin = qtplugin_factory(PyDMCheckbox)

def test_import_drawing_plugins():
    # Drawing plugins
    from ..widgets.drawing import (PyDMDrawingLine, PyDMDrawingRectangle, PyDMDrawingTriangle,
                                  PyDMDrawingEllipse, PyDMDrawingCircle, PyDMDrawingArc,
                                  PyDMDrawingPie, PyDMDrawingChord, PyDMDrawingImage,
                                  PyDMDrawingPolygon)

    PyDMDrawingImagePlugin = qtplugin_factory(PyDMDrawingImage)
    PyDMDrawingLinePlugin = qtplugin_factory(PyDMDrawingLine)
    PyDMDrawingRectanglePlugin = qtplugin_factory(PyDMDrawingRectangle)
    PyDMDrawingTrianglePlugin = qtplugin_factory(PyDMDrawingTriangle)
    PyDMDrawingEllipsePlugin = qtplugin_factory(PyDMDrawingEllipse)
    PyDMDrawingCirclePlugin = qtplugin_factory(PyDMDrawingCircle)
    PyDMDrawingArcPlugin = qtplugin_factory(PyDMDrawingArc)
    PyDMDrawingPiePlugin = qtplugin_factory(PyDMDrawingPie)
    PyDMDrawingChordPlugin = qtplugin_factory(PyDMDrawingChord)
    PyDMDrawingChordPlugin = qtplugin_factory(PyDMDrawingPolygon)

def test_import_embedded_display_plugin():
    # Embedded Display plugin
    from ..widgets.embedded_display import PyDMEmbeddedDisplay
    PyDMEmbeddedDisplayPlugin = qtplugin_factory(PyDMEmbeddedDisplay)

def test_import_frame_plugin():
    # Frame plugin
    from ..widgets.frame import PyDMFrame
    PyDMFramePlugin = qtplugin_factory(PyDMFrame, is_container=True)

def test_import_enum_button_plugin():
    # Enum Button plugin
    from ..widgets.enum_button import PyDMEnumButton
    PyDMEnumButtonPlugin = qtplugin_factory(PyDMEnumButton)

def test_import_combobox_plugin():
    # Enum Combobox plugin
    from ..widgets.enum_combo_box import PyDMEnumComboBox
    PyDMEnumComboBoxPlugin = qtplugin_factory(PyDMEnumComboBox)

def test_import_image_plugin():
    # Image plugin
    from ..widgets.image import PyDMImageView
    PyDMImageViewPlugin = qtplugin_factory(PyDMImageView)

def test_import_label_plugin():
    # Label plugin
    from ..widgets.label import PyDMLabel
    PyDMLabelPlugin = qtplugin_factory(PyDMLabel)

def test_import_line_edit_plugin():
    # Line Edit plugin
    from ..widgets.line_edit import PyDMLineEdit
    PyDMLineEditPlugin = qtplugin_factory(PyDMLineEdit)

def test_import_pushbutton_plugin():
    # Push Button plugin
    from ..widgets.pushbutton import PyDMPushButton
    PyDMPushButtonPlugin = qtplugin_factory(PyDMPushButton)

def test_import_related_display_plugin():
    # Related Display Button plugin
    from ..widgets.related_display_button import PyDMRelatedDisplayButton
    PyDMRelatedDisplayButtonPlugin = qtplugin_factory(PyDMRelatedDisplayButton)

def test_import_scale_indicator_plugin():
    # Scale Indicator plugin
    from ..widgets.scale import PyDMScaleIndicator
    PyDMScaleIndicatorPlugin = qtplugin_factory(PyDMScaleIndicator)

def test_import_shellcmd_plugin():
    # Shell Command plugin
    from ..widgets.shell_command import PyDMShellCommand
    PyDMShellCommandPlugin = qtplugin_factory(PyDMShellCommand)

def test_import_slider_plugin():
    # Slider plugin
    from ..widgets.slider import PyDMSlider
    PyDMSliderPlugin = qtplugin_factory(PyDMSlider)

def test_import_spinbox_plugin():
    # Spinbox plugin
    from ..widgets.spinbox import PyDMSpinbox
    PyDMSpinboxplugin = qtplugin_factory(PyDMSpinbox)

def test_import_symbol_plugin():
    # Symbol plugin
    from ..widgets.symbol import PyDMSymbol
    PyDMSymbolPlugin = qtplugin_factory(PyDMSymbol)

def test_import_waveform_table_plugin():
    # Waveform Table plugin
    from ..widgets.waveformtable import PyDMWaveformTable
    PyDMWaveformTablePlugin = qtplugin_factory(PyDMWaveformTable)

def test_import_timeplot_plugin():
    # Time Plot plugin
    from ..widgets.timeplot import PyDMTimePlot
    from ..widgets.timeplot_curve_editor import TimePlotCurveEditorDialog
    # Time Plot plugin
    PyDMTimePlotPlugin = qtplugin_factory(
                                PyDMTimePlot, TimePlotCurveEditorDialog)

def test_import_waveformplot_plugin():
    # Time Plot plugin
    from ..widgets.waveformplot import PyDMWaveformPlot
    from ..widgets.waveformplot_curve_editor import WaveformPlotCurveEditorDialog
    # Waveform Plot plugin
    PyDMWaveformPlotPlugin = qtplugin_factory(
                                PyDMWaveformPlot, WaveformPlotCurveEditorDialog)

def test_import_scatterplot_plugin():
    from ..widgets.scatterplot import PyDMScatterPlot
    from ..widgets.scatterplot_curve_editor import ScatterPlotCurveEditorDialog
    # Scatter Plot plugin
    PyDMScatterPlotPlugin = qtplugin_factory(
                                PyDMScatterPlot, ScatterPlotCurveEditorDialog)
                                
def test_import_tab_widget_plugin():
    # Symbol plugin
    from ..widgets.tab_bar import PyDMTabWidget
    from ..widgets.tab_bar_qtplugin import TabWidgetPlugin
    PyDMTabWidgetPlugin = TabWidgetPlugin()



from pyqtgraph.graphicsItems.ViewBox.ViewBoxMenu import ViewBoxMenu
from qtpy.QtCore import QCoreApplication, Signal
from qtpy.QtWidgets import QAction


class MultiAxisViewBoxMenu(ViewBoxMenu):
    """
    MultiAxisViewBoxMenu is a PyQtGraph ViewBoxMenu subclass to be used with MultiAxisViewBox. It can override
    the menu functionality to ensure that any selected option propagates through to each ViewBox in the plot.

    Parameters
    ----------
    view: ViewBox
        The view box this menu is associated with
    """

    # A signal indicating that the user has changed the mouse mode (left click panning vs. zooming)
    sigMouseModeChanged = Signal(object)
    # A signal indicating the user wants the default x and y axis ranges restored on this plot
    sigRestoreRanges = Signal()
    # A signal to set autorange for every view box on the plot
    sigSetAutorange = Signal(bool, bool)
    # A signal for updating the x autorange value
    sigXAutoRangeChanged = Signal(object)

    def __init__(self, view):
        super(MultiAxisViewBoxMenu, self).__init__(view)
        self.restoreRangesAction = QAction(QCoreApplication.translate("ViewBox", "Restore default X/Y ranges"), self)
        self.restoreRangesAction.triggered.connect(self.restoreRanges)

        # Insert/remove/insert because there is no QMenu function for inserting based on index
        self.insertAction(self.viewAll, self.restoreRangesAction)
        self.removeAction(self.viewAll)
        self.insertAction(self.restoreRangesAction, self.viewAll)

    def set3ButtonMode(self):
        """ Change the mouse left-click functionality to pan the plot """
        super(MultiAxisViewBoxMenu, self).set3ButtonMode()
        self.sigMouseModeChanged.emit('pan')

    def set1ButtonMode(self):
        """ Change the mouse left-click functionality to zoom in on the plot """
        super(MultiAxisViewBoxMenu, self).set1ButtonMode()
        self.sigMouseModeChanged.emit('rect')

    def xAutoClicked(self):
        """ Update the auto-range value for each view box """
        super().xAutoClicked()
        val = self.ctrl[0].autoPercentSpin.value() * 0.01
        self.sigXAutoRangeChanged.emit(val)

    def xManualClicked(self):
        """ Disable x auto-range for each view box """
        super().xManualClicked()
        self.sigXAutoRangeChanged.emit(False)

    def autoRange(self):
        """ Sets autorange to True for all elements on the plot """
        self.sigSetAutorange.emit(True, True)

    def restoreRanges(self):
        """ Restore the original x and y axis ranges for this plot """
        self.sigRestoreRanges.emit()



import math
import logging
from qtpy.QtCore import QTimer, Property, QEvent
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QLabel, QMessageBox, QApplication

from .base import PyDMPrimitiveWidget, get_icon_file
from ..utilities import is_qt_designer

logger = logging.getLogger(__name__)


FAST_TIMER_INTERVAL = 1000  # 1 second
SLOW_TIMER_INTERVAL = 10000  # 10 seconds


class PyDMTerminator(QLabel, PyDMPrimitiveWidget):
    """
    A watchdog widget to close a window after X seconds of inactivity.
    """
    designer_icon = QIcon(get_icon_file("terminator.png"))

    def __init__(self, parent=None, timeout=60, *args, **kwargs):
        super(PyDMTerminator, self).__init__(parent=parent, *args, **kwargs)
        self.setText("")
        self._hook_setup = False
        self._timeout = 60
        self._time_rem_ms = 0

        self._timer = QTimer()
        self._timer.timeout.connect(self.handle_timeout)
        self._timer.setSingleShot(True)

        self._window = None

        if timeout and timeout > 0:
            self.timeout = timeout
        else:
            self._reset()

        self._setup_activity_hook()
        self._update_label()

    def _find_window(self):
        """
        Finds the first window available starting from this widget's parent

        Returns
        -------
        QWidget
        """
        # check buffer
        if self._window:
            return self._window
        # go fish
        w = self.parent()
        while w is not None:
            if w.isWindow():
                return w
            w = w.parent()

        # we couldn't find it
        return None

    def _setup_activity_hook(self):
        if is_qt_designer():
            return
        logger.debug('Setup Hook')
        if self._hook_setup:
            logger.debug('Setup Hook Already there')
            return
        self._window = self._find_window()
        logger.debug('Install event filter at window')

        # We must install the event filter in the application otherwise
        # it won't stop when typing or moving over other widgets or even
        # the PyDM main window if in use.
        QApplication.instance().installEventFilter(self)
        self._hook_setup = True

    def eventFilter(self, obj, ev):
        if ev.type() in (QEvent.MouseMove, QEvent.KeyPress, QEvent.KeyRelease):
            self.reset()
        return super(PyDMTerminator, self).eventFilter(obj, ev)

    def reset(self):
        if self._time_rem_ms != self._timeout * 1000:
            self._time_rem_ms = self._timeout * 1000
            self._update_label()
        self.stop()
        self.start()

    def start(self):
        if is_qt_designer():
            return
        interval = SLOW_TIMER_INTERVAL
        if self._time_rem_ms < 60*1000:
            interval = FAST_TIMER_INTERVAL
        self._timer.setInterval(interval)

        if not self._timer.isActive():
            self._timer.start()

    def stop(self):
        if self._timer.isActive():
            self._timer.stop()

    def _get_time_text(self, value):
        """
        Converts value in seconds into a text for days, hours, minutes and
        seconds remaining.

        Parameters
        ----------
        value : int
            The value in seconds to be converted

        Returns
        -------
        str
        """
        def time_msg(unit, val):
            return "{} {}{}".format(val, unit, "s" if val > 1 else "")

        units = ["day", "hour", "minute", "second"]
        scale = [86400, 3600, 60, 1]

        values = [0, 0, 0, 0]
        rem = value
        for idx, sc in enumerate(scale):
            val_scaled, rem = int(rem//sc), rem % sc
            if val_scaled >= 1:
                val_scaled = math.ceil(val_scaled+(rem/sc))
                values[idx] = val_scaled
                break
            values[idx] = val_scaled

        time_items = []
        for idx, un in enumerate(units):
            v = values[idx]
            if v > 0:
                time_items.append(time_msg(un, v))

        return ", ".join(time_items)

    def _update_label(self):
        """Updates the label text with the remaining time."""
        rem_time_s = self._time_rem_ms/1000.0
        text = self._get_time_text(rem_time_s)
        self.setText("This screen will close in {}.".format(text))

    @Property(int)
    def timeout(self):
        """
        Timeout in seconds.

        Returns
        -------
        int
        """
        return self._timeout

    @timeout.setter
    def timeout(self, seconds):
        self.stop()
        if seconds and seconds > 0:
            self._timeout = seconds
            self.reset()

    def handle_timeout(self):
        """
        Handles the timeout event for the timer.
        Decreases the time remaining counter until 0 and when it is time,
        cleans up the event filter and closes the window.
        """
        if is_qt_designer():
            return
        # Decrease remaining time
        self._time_rem_ms -= self._timer.interval()
        # Update screen label with new remaining time
        self._update_label()

        if self._time_rem_ms > 0:
            self.start()
            return
        QApplication.instance().removeEventFilter(self)

        if self._window:
            logger.debug('Time to close the window')
            self._window.close()
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Warning)
            msg.setText(
                "Your window was closed due to inactivity for {}.".format(
                    self._get_time_text(self.timeout)
                )
            )
            msg.setStandardButtons(QMessageBox.Ok)
            msg.setDefaultButton(QMessageBox.Ok)
            msg.exec_()



import logging

from qtpy.QtWidgets import QWidget

logger = logging.getLogger(__name__)


def _change_connection_status(widget, status):
    """
    Connect or disconnect the inner channels of widgets on the
    given widget based on the status parameter.

    Parameters
    ----------
    widget : QWidget
        The widget which will be iterated over for channel connection.

    status : bool
        If True, will call connect on the channels otherwise it will call
        disconnect.
    """
    widgets = [widget]
    widgets.extend(widget.findChildren(QWidget))
    for child_widget in widgets:
        try:
            if hasattr(child_widget, 'channels'):
                if child_widget.channels() is None:
                    continue
                for channel in child_widget.channels():
                    if channel is None:
                        continue
                    if status:
                        channel.connect()
                    else:
                        channel.disconnect()
        except NameError:
            continue

def establish_widget_connections(widget):
    """
    Connect the inner channels of widgets on the given widget.

    Parameters
    ----------
    widget : QWidget
        The widget which will be iterated over for channel connection.
    """
    _change_connection_status(widget, True)


def close_widget_connections(widget):
    """
    Disconnect the inner channels of widgets on the given widget.

    Parameters
    ----------
    widget : QWidget
        The widget which will be iterated over for channel disconnection.
    """
    _change_connection_status(widget, False)



import os
from os import path
from qtpy.QtWidgets import (QApplication, QMainWindow, QFileDialog,
                            QWidget, QAction, QMessageBox)
from qtpy.QtCore import Qt, QTimer, Slot, QSize, QLibraryInfo
from .utilities import (IconFont, find_file, establish_widget_connections,
                        close_widget_connections)
from .pydm_ui import Ui_MainWindow
from .display import Display, ScreenTarget, load_file
from .connection_inspector import ConnectionInspector
from .about_pydm import AboutWindow
from . import data_plugins
from . import tools
from .widgets.rules import register_widget_rules, unregister_widget_rules
from . import config
import subprocess
import platform
import logging

logger = logging.getLogger(__name__)


class PyDMMainWindow(QMainWindow):

    def __init__(self, parent=None, hide_nav_bar=False, hide_menu_bar=False, hide_status_bar=False):
        super(PyDMMainWindow, self).__init__(parent)
        self.app = QApplication.instance()
        self.font_factor = 1
        self.iconFont = IconFont()
        self._display_widget = None
        self._showing_file_path_in_title_bar = False

        self._saved_menu_geometry = None
        self._saved_menu_height = None
        self._new_widget_size = None

        self.default_font_size = QApplication.instance().font().pointSizeF()

        self.home_file = None
        self.home_widget = None

        self.designer_path = None

        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self.ui.navbar.setIconSize(QSize(24, 24))
        self.ui.navbar.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
        self.ui.actionHome.triggered.connect(self.home_triggered)
        self.ui.actionHome.setIcon(self.iconFont.icon("home"))
        self.ui.actionBack.triggered.connect(self.back_triggered)
        self.ui.actionBack.setIcon(self.iconFont.icon("angle-left"))
        self.ui.actionForward.triggered.connect(self.forward_triggered)
        self.ui.actionForward.setIcon(self.iconFont.icon("angle-right"))
        self.ui.actionEdit_in_Designer.triggered.connect(self.edit_in_designer)
        self.ui.actionOpen_File.triggered.connect(self.open_file_action)
        self.ui.actionReload_Display.triggered.connect(self.reload_display)
        self.ui.actionIncrease_Font_Size.triggered.connect(self.increase_font_size)
        self.ui.actionDecrease_Font_Size.triggered.connect(self.decrease_font_size)
        self.ui.actionDefault_Font_Size.triggered.connect(self.reset_font_size)
        self.ui.actionEnter_Fullscreen.triggered.connect(self.enter_fullscreen)
        self.ui.actionShow_File_Path_in_Title_Bar.triggered.connect(self.toggle_file_path_in_title_bar)
        self.ui.actionShow_Navigation_Bar.triggered.connect(self.toggle_nav_bar)
        self.ui.actionShow_Menu_Bar.triggered.connect(self.toggle_menu_bar)
        self.ui.actionShow_Status_Bar.triggered.connect(self.toggle_status_bar)
        self.ui.actionShow_Connections.triggered.connect(self.show_connections)
        self.ui.actionAbout_PyDM.triggered.connect(self.show_about_window)
        self.ui.actionLoadTool.triggered.connect(self.load_tool)
        self.ui.actionLoadTool.setIcon(self.iconFont.icon("rocket"))
        self.ui.actionQuit.triggered.connect(self.quit_main_window)

        if hide_nav_bar:
            self.toggle_nav_bar(False)
        if hide_menu_bar:
            # Toggle the menu bar via the QAction so that the menu item
            # stays in sync with menu visibility.
            self.ui.actionShow_Menu_Bar.activate(QAction.Trigger)
        if hide_status_bar:
            self.toggle_status_bar(False)
        #Try to find the designer binary.
        self.ui.actionEdit_in_Designer.setEnabled(False)

        possible_designer_bin_paths = (QLibraryInfo.location(QLibraryInfo.BinariesPath), QLibraryInfo.location(QLibraryInfo.LibraryExecutablesPath))
        for bin_path in possible_designer_bin_paths:
            if platform.system() == 'Darwin':
                designer_path = os.path.join(bin_path, 'Designer.app/Contents/MacOS/Designer')
            elif platform.system() == 'Linux':
                designer_path = os.path.join(bin_path, 'designer')
            else:
                designer_path = os.path.join(bin_path, 'designer.exe')
            if os.path.isfile(designer_path):
                self.designer_path = designer_path
                break

        self.update_tools_menu()
        self.enable_disable_navigation()

    def display_widget(self):
        return self._display_widget

    def set_display_widget(self, new_widget):
        if new_widget == self._display_widget:
            return
        self.clear_display_widget()
        if not new_widget.layout():
            new_widget.setMinimumSize(new_widget.size())
        self._new_widget_size = new_widget.size()
        new_widget.setVisible(True)
        self._display_widget = new_widget
        self.setCentralWidget(self._display_widget)
        self.enable_disable_navigation()
        self.update_window_title()
        # Resizing to the new widget's dimensions needs to be
        # done on the event loop for some reason - you can't
        # just do it here.
        QTimer.singleShot(0, self.resizeForNewDisplayWidget)

    def clear_display_widget(self):
        if self._display_widget is not None:
            close_widget_connections(self._display_widget)
            unregister_widget_rules(self._display_widget)
            self._display_widget.setVisible(False)
            self._display_widget.setParent(None)
            self.ui.actionEdit_in_Designer.setEnabled(False)

    def handle_open_file_error(self, filename, error):
        self.statusBar().showMessage("Cannot open file: '{0}', reason: '{1}'...".format(filename, error), 5000)

    def back_triggered(self):
        new_process = QApplication.keyboardModifiers() == Qt.ShiftModifier
        self.back(open_in_new_process=new_process)

    def back(self, open_in_new_process=False):
        curr_display = self.display_widget()
        prev_display = None
        if curr_display:
            prev_display = curr_display.previous_display

        if not prev_display:
            logger.error('No display history to execute back navigation.')
            return

        if open_in_new_process:
            load_file(prev_display.loaded_file(),
                      macros=prev_display.macros,
                      args=prev_display.args,
                      target=ScreenTarget.NEW_PROCESS)
        else:
            prev_display.next_display = curr_display
            establish_widget_connections(prev_display)
            register_widget_rules(prev_display)
            self.set_display_widget(prev_display)

    def forward_triggered(self):
        new_process = QApplication.keyboardModifiers() == Qt.ShiftModifier
        self.forward(open_in_new_process=new_process)

    def forward(self, open_in_new_process=False):
        curr_display = self.display_widget()
        next_display = None
        if curr_display:
            next_display = curr_display.next_display

        if not next_display:
            logger.error('No display history to execute forward navigation.')
            return

        if open_in_new_process:
            load_file(next_display.loaded_file(),
                      macros=next_display.macros,
                      args=next_display.args,
                      target=ScreenTarget.NEW_PROCESS)
        else:
            establish_widget_connections(next_display)
            register_widget_rules(next_display)
            next_display.previous_display = curr_display
            self.set_display_widget(next_display)

    def home_triggered(self):
        new_process = QApplication.keyboardModifiers() == Qt.ShiftModifier
        self.home(open_in_new_process=new_process)

    def home(self, open_in_new_process=False):
        if self.home_widget is None:
            return

        if open_in_new_process:
            fname = self.home_widget.loaded_file()
            macros = self.home_widget.macros()
            args = self.home_widget.args()
            load_file(fname,
                      macros=macros,
                      args=args,
                      target=ScreenTarget.NEW_PROCESS)
        else:
            self.set_display_widget(self.home_widget)

    def enable_disable_navigation(self):
        w = self.display_widget()
        if not isinstance(w, Display):
            # We can't do much if it is not a Display and we don't have the
            # previous_display and next_display properties since we don't
            # have the navigation stack set.
            nav_stack_methods = hasattr(w, 'previous_display') \
                                and hasattr(w, 'next_display')
            if not nav_stack_methods:
                return
        if not w:
            self.ui.actionBack.setDisabled(True)
            self.ui.actionForward.setDisabled(True)
            return
        self.ui.actionBack.setDisabled(w.previous_display is None)
        self.ui.actionForward.setDisabled(w.next_display is None)

    def current_file(self):
        return self.display_widget().loaded_file()

    def update_window_title(self):
        if self.showing_file_path_in_title_bar:
            title = self.current_file()
        else:
            title = self.display_widget().windowTitle()
        title += " - PyDM"
        if data_plugins.is_read_only():
            title += " [Read Only Mode]"
        self.setWindowTitle(title)

    @property
    def showing_file_path_in_title_bar(self):
        return self._showing_file_path_in_title_bar

    @showing_file_path_in_title_bar.setter
    def showing_file_path_in_title_bar(self, checked):
        self._showing_file_path_in_title_bar = checked
        self.update_window_title()

    @Slot(bool)
    def toggle_file_path_in_title_bar(self, checked):
        self.showing_file_path_in_title_bar = checked

    @Slot(bool)
    def toggle_nav_bar(self, checked):
        self.ui.navbar.setHidden(not checked)

    @Slot(bool)
    def toggle_menu_bar(self, checked=None):
        if checked is None:
            checked = not self.ui.menubar.isVisible()
        # Crazy hack: we can't just do menubar.setVisible(), because that
        # will disable all the QActions and their keyboard shortcuts when
        # we hide the menu.  So instead, we set it to a height of 0 to hide
        # it, and then restore the previous height value to show it again.
        if checked:
            self.ui.menubar.restoreGeometry(self._saved_menu_geometry)
            self.ui.menubar.setFixedHeight(self._saved_menu_height)
        else:
            self._saved_menu_geometry = self.ui.menubar.saveGeometry()
            self._saved_menu_height = self.ui.menubar.height()
            self.ui.menubar.setFixedHeight(0)

    @Slot(bool)
    def toggle_status_bar(self, checked):
        self.ui.statusbar.setHidden(not checked)

    def get_files_in_display(self):
        try:
            curr_file = self.current_file()
        except IndexError:
            logger.error("The display manager does not have a display loaded.")
            return None, None

        _, extension = path.splitext(curr_file)
        if extension == '.ui':
            return self.current_file(), None
        else:
            central_widget = self.centralWidget() if isinstance(self.centralWidget(), Display) else None
            if central_widget is not None:
                ui_file = central_widget.ui_filepath()
            return ui_file, self.current_file()

    @Slot(bool)
    def edit_in_designer(self, checked):

        def open_editor_ui(fname):
            if self.designer_path is None or fname is None or fname == "":
                return
            self.statusBar().showMessage("Launching '{0}' in Qt Designer...".format(fname), 5000)
            _ = subprocess.Popen('{0} "{1}"'.format(self.designer_path, fname), shell=True)

        def open_editor_generic(fname):
            if platform.system() == "Darwin":
                subprocess.call(('open', fname))
            elif platform.system() == "Linux":
                subprocess.call(('xdg-open', fname))
            elif platform.system() == "Windows":
                os.startfile(fname)

        ui_file, py_file = self.get_files_in_display()
        if ui_file is not None and ui_file != "":
            open_editor_ui(fname=ui_file)
        if py_file is not None and py_file != "":
            open_editor_generic(fname=py_file)

    @Slot(bool)
    def open_file_action(self, checked):
        modifiers = QApplication.keyboardModifiers()
        try:
            curr_file = self.current_file()
            folder = os.path.dirname(curr_file)
        except Exception:
            folder = os.getcwd()

        filename = QFileDialog.getOpenFileName(self, 'Open File...', folder, 'PyDM Display Files (*.ui *.py)')
        filename = filename[0] if isinstance(filename, (list, tuple)) else filename

        if filename:
            filename = str(filename)
            try:
                if modifiers == Qt.ShiftModifier:
                    target = ScreenTarget.NEW_PROCESS
                else:
                    target = None
                self.open(filename, target=target)

            except (IOError, OSError, ValueError, ImportError) as e:
                self.handle_open_file_error(filename, e)

    def open(self, filename, macros=None, args=None, target=None):
        if not os.path.isabs(filename):
            base_path = None
            curr_display = self.display_widget()
            if curr_display:
                base_path = os.path.dirname(curr_display.loaded_file())
            filename = find_file(filename, base_path=base_path)
        new_widget = load_file(filename,
                               macros=macros,
                               args=args,
                               target=target)
        if new_widget:
            if self.home_widget is None:
                self.home_widget = new_widget
            display_widget = self.display_widget()
            if display_widget:
                new_widget.previous_display = display_widget
            self.set_display_widget(new_widget)
            ui_file, py_file = self.get_files_in_display()
            editors = []
            if ui_file:
                editors.append("Designer")
            if py_file:
                editors.append("Text Editor")
            edit_in_text = "Open in {}".format(' and '.join(editors))
            self.ui.actionEdit_in_Designer.setText(edit_in_text)
            if (self.designer_path and ui_file) or (py_file and not ui_file):
                self.ui.actionEdit_in_Designer.setEnabled(True)
        return new_widget

    def load_tool(self, checked):
        try:
            curr_dir = os.path.dirname(self.current_file())
        except Exception:
            curr_dir = os.getcwd()
            logger.error("The display manager does not have a display loaded. Suggesting current work directory.")
        filename = QFileDialog.getOpenFileName(self, 'Load tool...', curr_dir, 'PyDM External Tool Files (*_tool.py)')
        filename = filename[0] if isinstance(filename, (list, tuple)) else filename

        if filename:
            filename = str(filename)
            tools.install_external_tool(filename)
            self.update_tools_menu()

    def update_tools_menu(self):
        """
        Update the Main Window Tools menu.
        """
        kwargs = {'channels': None, 'sender': self}
        tools.assemble_tools_menu(self.ui.menuTools,
                                  clear_menu=True,
                                  **kwargs)

        self.ui.menuTools.addSeparator()
        self.ui.menuTools.addAction(self.ui.actionLoadTool)

    @Slot(bool)
    def reload_display(self, checked):
        curr_display = self.display_widget()
        if not curr_display:
            logger.error("The display manager does not have a display loaded.")
            return

        prev_display = curr_display.previous_display
        next_display = curr_display.next_display

        macros = curr_display.macros()
        args = curr_display.args()
        loaded_file = curr_display.loaded_file()

        self.statusBar().showMessage(
            "Reloading '{0}'...".format(self.current_file()), 5000)
        new_widget = self.open(loaded_file, macros=macros, args=args)
        new_widget.previous_display = prev_display
        new_widget.next_display = next_display

    @Slot(bool)
    def increase_font_size(self, checked):
        old_factor = self.font_factor
        self.font_factor += 0.1
        self.set_font_size(old_factor, self.font_factor)

    @Slot(bool)
    def decrease_font_size(self, checked):
        old_factor = self.font_factor
        self.font_factor -= 0.1
        self.set_font_size(old_factor, self.font_factor)

    @Slot(bool)
    def reset_font_size(self, checked):
        old_factor = self.font_factor
        self.font_factor = 1
        self.set_font_size(old_factor, self.font_factor)

    def set_font_size(self, old, new):
        current_font = self.app.font()
        current_font.setPointSizeF(current_font.pointSizeF()/old*new)
        QApplication.instance().setFont(current_font)

        for w in self.app.allWidgets():
            w_c_f = w.font()
            w_c_f.setPointSizeF(w_c_f.pointSizeF()/old*new)
            w.setFont(w_c_f)

        QTimer.singleShot(0, self.resizeForNewDisplayWidget)

    @Slot(bool)
    def enter_fullscreen(self, checked=False):
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    @Slot(bool)
    def show_connections(self, checked):
        c = ConnectionInspector(self)
        c.show()

    @Slot(bool)
    def show_about_window(self, checked):
        a = AboutWindow(self)
        a.show()

    def resizeForNewDisplayWidget(self):
        if not self.isFullScreen():
            self.resize(self._new_widget_size)

    def closeEvent(self, event):
        event.ignore()

        def do_close():
            self.clear_display_widget()
            event.accept()

        main_windows = [w for w in self.app.topLevelWidgets() if isinstance(w, QMainWindow)]
        if len(main_windows) == 1 and main_windows[0] == self:
            self._confirm_quit(callback=do_close)
        else:
            do_close()

    @Slot(bool)
    def quit_main_window(self, checked):
        self._confirm_quit(callback=self.app.quit)

    def _confirm_quit(self, callback):
        if not config.CONFIRM_QUIT:
            callback()
        else:
            quit_message = QMessageBox.question(
                self, 'Quitting Application', 'Exit Application?',
                QMessageBox.Yes | QMessageBox.No)
            if quit_message == QMessageBox.Yes:
                callback()



import logging
logger = logging.getLogger(__name__)

from qtpy.QtWidgets import QFrame, QLabel, QSlider, QVBoxLayout, QHBoxLayout, QSizePolicy, QWidget
from qtpy.QtCore import Qt, Signal, Slot, Property
from .base import PyDMWritableWidget, TextFormatter
import numpy as np


class PyDMSlider(QFrame, TextFormatter, PyDMWritableWidget):
    """
    A QSlider with support for Channels and more from PyDM.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the Label
    init_channel : str, optional
        The channel to be used by the widget.
    """
    actionTriggered = Signal(int)
    rangeChanged = Signal(float, float)
    sliderMoved = Signal(float)
    sliderPressed = Signal()
    sliderReleased = Signal()
    valueChanged = Signal(float)

    def __init__(self, parent=None, init_channel=None):
        QFrame.__init__(self, parent)
        PyDMWritableWidget.__init__(self, init_channel=init_channel)
        self.alarmSensitiveContent = True
        self.alarmSensitiveBorder = False
        # Internal values for properties
        self._ignore_mouse_wheel = True
        self._show_limit_labels = True
        self._show_value_label = True
        self._user_defined_limits = False
        self._needs_limit_info = True
        self._minimum = None
        self._maximum = None
        self._user_minimum = -10.0
        self._user_maximum = 10.0
        self._num_steps = 101
        self._orientation = Qt.Horizontal
        # Set up all the internal widgets that make up a PyDMSlider.
        # We'll add all these things to layouts when we call setup_widgets_for_orientation
        label_size_policy = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        self.low_lim_label = QLabel(self)
        self.low_lim_label.setObjectName("lowLimLabel")
        self.low_lim_label.setSizePolicy(label_size_policy)
        self.low_lim_label.setAlignment(Qt.AlignLeft | Qt.AlignTrailing | Qt.AlignVCenter)
        self.value_label = QLabel(self)
        self.value_label.setObjectName("valueLabel")
        self.value_label.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        self.high_lim_label = QLabel(self)
        self.high_lim_label.setObjectName("highLimLabel")
        self.high_lim_label.setSizePolicy(label_size_policy)
        self.high_lim_label.setAlignment(Qt.AlignRight | Qt.AlignTrailing | Qt.AlignVCenter)
        self._slider = QSlider(parent=self)
        self._slider.setOrientation(Qt.Horizontal)

        self._orig_wheel_event = self._slider.wheelEvent
        self._slider.wheelEvent = self.wheelEvent

        self._slider.sliderMoved.connect(self.internal_slider_moved)
        self._slider.sliderPressed.connect(self.internal_slider_pressed)
        self._slider.sliderReleased.connect(self.internal_slider_released)
        self._slider.valueChanged.connect(self.internal_slider_value_changed)
        # self.vertical_layout.addWidget(self._slider)
        # Other internal variables and final setup steps
        self._slider_position_to_value_map = None
        self._mute_internal_slider_changes = False
        self.setup_widgets_for_orientation(self._orientation)
        self.reset_slider_limits()

    def wheelEvent(self, e):
        #We specifically want to ignore mouse wheel events.
        if self._ignore_mouse_wheel:
            e.ignore()
        else:
            super(PyDMSlider, self).wheelEvent(e)
        return

    def init_for_designer(self):
        """
        Method called after the constructor to tweak configurations for
        when using the widget with the Qt Designer
        """
        self.value = 0.0

    @Property(Qt.Orientation)
    def orientation(self):
        """
        The slider orientation (Horizontal or Vertical)

        Returns
        -------
        int
            Qt.Horizontal or Qt.Vertical
        """
        return self._orientation

    @orientation.setter
    def orientation(self, new_orientation):
        """
        The slider orientation (Horizontal or Vertical)

        Parameters
        ----------
        new_orientation : int
            Qt.Horizontal or Qt.Vertical
        """
        self._orientation = new_orientation
        self.setup_widgets_for_orientation(new_orientation)

    def setup_widgets_for_orientation(self, new_orientation):
        """
        Reconstruct the widget given the orientation.

        Parameters
        ----------
        new_orientation : int
            Qt.Horizontal or Qt.Vertical
        """
        if new_orientation not in (Qt.Horizontal, Qt.Vertical):
            logger.error("Invalid orientation '{0}'. The existing layout will not change.".format(new_orientation))
            return

        layout = None
        if new_orientation == Qt.Horizontal:
            layout = QVBoxLayout()
            layout.setContentsMargins(4, 0, 4, 4)
            label_layout = QHBoxLayout()
            label_layout.addWidget(self.low_lim_label)
            label_layout.addStretch(0)
            label_layout.addWidget(self.value_label)
            label_layout.addStretch(0)
            label_layout.addWidget(self.high_lim_label)
            layout.addLayout(label_layout)
            self._slider.setOrientation(new_orientation)
            layout.addWidget(self._slider)
        elif new_orientation == Qt.Vertical:
            layout = QHBoxLayout()
            layout.setContentsMargins(0, 4, 4, 4)
            label_layout = QVBoxLayout()
            label_layout.addWidget(self.high_lim_label)
            label_layout.addStretch(0)
            label_layout.addWidget(self.value_label)
            label_layout.addStretch(0)
            label_layout.addWidget(self.low_lim_label)
            layout.addLayout(label_layout)
            self._slider.setOrientation(new_orientation)
            layout.addWidget(self._slider)

        if self.layout() is not None:
            # Trick to remove the existing layout by re-parenting it in an empty widget.
            QWidget().setLayout(self.layout())
        self.setLayout(layout)

    def update_labels(self):
        """
        Update the limits and value labels with the correct values.
        """
        def set_label(value, label_widget):
            if value is None:
                label_widget.setText("")
            else:
                label_widget.setText(self.format_string.format(value))

        set_label(self.minimum, self.low_lim_label)
        set_label(self.maximum, self.high_lim_label)
        set_label(self.value, self.value_label)

    def reset_slider_limits(self):
        """
        Reset the limits and adjust the labels properly for the slider.
        """
        logger.debug("Running reset_slider_limits.")
        if self.minimum is None or self.maximum is None:
            self._needs_limit_info = True
            logger.debug("Need both limits before reset_slider_limits can work.")
            self.set_enable_state()
            return
        logger.debug("Has both limits, proceeding.")
        self._needs_limit_info = False
        self._slider.setMinimum(0)
        self._slider.setMaximum(self._num_steps - 1)
        self._slider.setSingleStep(1)
        self._slider.setPageStep(10)
        self._slider_position_to_value_map = np.linspace(self.minimum, self.maximum, num=self._num_steps)
        self.update_labels()
        self.set_slider_to_closest_value(self.value)
        self.rangeChanged.emit(self.minimum, self.maximum)
        self.set_enable_state()

    def find_closest_slider_position_to_value(self, val):
        """
        Find and returns the index for the closest position on the slider
        for a given value.

        Parameters
        ----------
        val : float

        Returns
        -------
        int
        """
        diff = abs(self._slider_position_to_value_map - float(val))
        logger.debug("The closest value to %f is: %f", val, self._slider_position_to_value_map[np.argmin(diff)])
        return np.argmin(diff)

    def set_slider_to_closest_value(self, val):
        """
        Set the value for the slider according to a given value.

        Parameters
        ----------
        val : float
        """
        if val is None or self._needs_limit_info:
            logger.debug("Not setting slider to closest value because we need limits.")
            return
        # When we set the slider to the closest value, it may end up at a slightly
        # different position than val (if val is not in self._slider_position_to_value_map)
        # We don't want that slight difference to get broacast out and put the channel
        # somewhere new.    For example, if the slider can only be at 0.4 or 0.5, but a
        # new value comes in of 0.45, its more important to keep the 0.45 than to change
        # it to where the slider gets set.  Therefore, we mute the internal slider changes
        # so that its valueChanged signal doesn't cause us to emit a signal to PyDM to change
        # the value of the channel.
        logger.debug("Setting slider to closest value.")
        self._mute_internal_slider_changes = True
        self._slider.setValue(self.find_closest_slider_position_to_value(val))
        self._mute_internal_slider_changes = False

    def connection_changed(self, connected):
        """
        Callback invoked when the connection state of the Channel is changed.
        This callback acts on the connection state to enable/disable the widget
        and also trigger the change on alarm severity to ALARM_DISCONNECTED.

        Parameters
        ----------
        connected : int
            When this value is 0 the channel is disconnected, 1 otherwise.
        """
        super(PyDMSlider, self).connection_changed(connected)
        self.set_enable_state()

    def write_access_changed(self, new_write_access):
        """
        Callback invoked when the Channel has new write access value.
        This callback calls check_enable_state so it can act on the widget
        enabling or disabling it accordingly

        Parameters
        ----------
        new_write_access : bool
            True if write operations to the channel are allowed.
        """
        super(PyDMSlider, self).write_access_changed(new_write_access)
        self.set_enable_state()

    def value_changed(self, new_val):
        """
        Callback invoked when the Channel value is changed.

        Parameters
        ----------
        new_val : int or float
            The new value from the channel.
        """
        PyDMWritableWidget.value_changed(self, new_val)
        if hasattr(self, "value_label"):
            logger.debug("Setting text for value label.")
            self.value_label.setText(self.format_string.format(self.value))
        if not self._slider.isSliderDown():
            self.set_slider_to_closest_value(self.value)
        self.update_format_string()

    def ctrl_limit_changed(self, which, new_limit):
        """
        Callback invoked when the Channel receives new control limit
        values.

        Parameters
        ----------
        which : str
            Which control limit was changed. "UPPER" or "LOWER"
        new_limit : float
            New value for the control limit
        """
        logger.debug("%s limit changed to %f", which, new_limit)
        PyDMWritableWidget.ctrl_limit_changed(self, which, new_limit)
        if not self.userDefinedLimits:
            self.reset_slider_limits()

    def update_format_string(self):
        """
        Reconstruct the format string to be used when representing the
        output value.

        Returns
        -------
        format_string : str
            The format string to be used including or not the precision
            and unit
        """
        fs = super(PyDMSlider, self).update_format_string()
        self.update_labels()
        return fs

    def set_enable_state(self):
        """
        Determines wether or not the widget must be enabled or not depending
        on the write access, connection state and presence of limits information
        """
        # Even though by documentation disabling parent QFrame (self), should disable internal
        # slider, in practice it still remains interactive (PyQt 5.12.1). Disabling explicitly, solves
        # the problem.
        should_enable = self._write_access and self._connected and not self._needs_limit_info
        self.setEnabled(should_enable)
        self._slider.setEnabled(should_enable)

    @Slot(int)
    def internal_slider_action_triggered(self, action):
        self.actionTriggered.emit(action)

    @Slot(int)
    def internal_slider_moved(self, val):
        """
        Method invoked when the slider is moved.

        Parameters
        ----------
        val : float
        """
        self.sliderMoved.emit(self.value)

    @Slot()
    def internal_slider_pressed(self):
        """
        Method invoked when the slider is pressed
        """
        self.sliderPressed.emit()

    @Slot()
    def internal_slider_released(self):
        """
        Method invoked when the slider is released
        """
        self.sliderReleased.emit()

    @Slot(int)
    def internal_slider_value_changed(self, val):
        """
        Method invoked when a new value is selected on the slider.
        This will cause the new value to be emitted to the signal
        unless `mute_internal_slider_changes` is True.

        Parameters
        ----------
        val : int
        """        
        # Avoid potential crash if limits are undefined
        if self._slider_position_to_value_map is None:
            return
        if not self._mute_internal_slider_changes:
            self.value = self._slider_position_to_value_map[val]
            self.send_value_signal[float].emit(self.value)

    @Property(bool)
    def tracking(self):
        """
        If tracking is enabled (the default), the slider emits new values
        while the slider is being dragged.  If tracking is disabled, it will
        only emit new values when the user releases the slider.  Tracking can
        cause PyDM to rapidly send new values to the channel.  If you are using
        the slider to control physical hardware, consider whether the device
        you want to control can handle large amounts of changes in a short
        timespan.
        """
        return self._slider.hasTracking()
        
    @tracking.setter
    def tracking(self, checked):
        self._slider.setTracking(checked)
    
    def hasTracking(self):
        """
        An alternative function to get the tracking property, to match what
        Qt provides for QSlider.
        """
        return self.tracking
    
    def setTracking(self, checked):
        """
        An alternative function to set the tracking property, to match what
        Qt provides for QSlider.
        """
        self.tracking = checked

    @Property(bool)
    def ignoreMouseWheel(self):
        """
        If true, the mouse wheel will not change the value of the slider.
        This is useful if you want to put sliders inside a scroll view, and
        don't want to accidentally change the slider as you are scrolling.
        """
        return self._ignore_mouse_wheel
    
    @ignoreMouseWheel.setter
    def ignoreMouseWheel(self, checked):
        self._ignore_mouse_wheel = checked
        if checked:
            self._slider.wheelEvent = self.wheelEvent
        else:
            self._slider.wheelEvent = self._orig_wheel_event

    @Property(bool)
    def showLimitLabels(self):
        """
        Whether or not the high and low limits should be displayed on the slider.

        Returns
        -------
        bool
        """
        return self._show_limit_labels

    @showLimitLabels.setter
    def showLimitLabels(self, checked):
        """
        Whether or not the high and low limits should be displayed on the slider.

        Parameters
        ----------
        checked : bool
        """
        self._show_limit_labels = checked
        if checked:
            self.low_lim_label.show()
            self.high_lim_label.show()
        else:
            self.low_lim_label.hide()
            self.high_lim_label.hide()

    @Property(bool)
    def showValueLabel(self):
        """
        Whether or not the current value should be displayed on the slider.

        Returns
        -------
        bool
        """
        return self._show_value_label

    @showValueLabel.setter
    def showValueLabel(self, checked):
        """
        Whether or not the current value should be displayed on the slider.

        Parameters
        ----------
        checked : bool
        """
        self._show_value_label = checked
        if checked:
            self.value_label.show()
        else:
            self.value_label.hide()

    @Property(QSlider.TickPosition)
    def tickPosition(self):
        """
        Where to draw tick marks for the slider.

        Returns
        -------
        QSlider.TickPosition
        """
        return self._slider.tickPosition()

    @tickPosition.setter
    def tickPosition(self, position):
        """
        Where to draw tick marks for the slider.

        Parameter
        ---------
        position : QSlider.TickPosition
        """
        self._slider.setTickPosition(position)

    @Property(bool)
    def userDefinedLimits(self):
        """
        Wether or not to use limits defined by the user and not from the
        channel

        Returns
        -------
        bool
        """
        return self._user_defined_limits

    @userDefinedLimits.setter
    def userDefinedLimits(self, user_defined_limits):
        """
        Wether or not to use limits defined by the user and not from the
        channel

        Parameters
        ----------
        user_defined_limits : bool
        """
        self._user_defined_limits = user_defined_limits
        self.reset_slider_limits()

    @Property(float)
    def userMinimum(self):
        """
        Lower user defined limit value

        Returns
        -------
        float
        """
        return self._user_minimum

    @userMinimum.setter
    def userMinimum(self, new_min):
        """
        Lower user defined limit value

        Parameters
        ----------
        new_min : float
        """
        self._user_minimum = float(new_min) if new_min is not None else None
        if self.userDefinedLimits:
            self.reset_slider_limits()

    @Property(float)
    def userMaximum(self):
        """
        Upper user defined limit value

        Returns
        -------
        float
        """
        return self._user_maximum

    @userMaximum.setter
    def userMaximum(self, new_max):
        """
        Upper user defined limit value

        Parameters
        ----------
        new_max : float
        """
        self._user_maximum = float(new_max) if new_max is not None else None
        if self.userDefinedLimits:
            self.reset_slider_limits()

    @property
    def minimum(self):
        """
        The current value being used for the lower limit

        Returns
        -------
        float
        """
        if self.userDefinedLimits:
            return self._user_minimum
        return self._lower_ctrl_limit

    @property
    def maximum(self):
        """
        The current value being used for the upper limit

        Returns
        -------
        float
        """
        if self.userDefinedLimits:
            return self._user_maximum
        return self._upper_ctrl_limit

    @Property(int)
    def num_steps(self):
        """
        The number of steps on the slider

        Returns
        -------
        int
        """
        return self._num_steps

    @num_steps.setter
    def num_steps(self, new_steps):
        """
        The number of steps on the slider

        Parameters
        ----------
        new_steps : int
        """
        self._num_steps = int(new_steps)
        self.reset_slider_limits()



import numpy as np
from pyqtgraph import BarGraphItem
from unittest import mock
from ...widgets.waveformplot import PyDMWaveformPlot, WaveformCurveItem


@mock.patch('pydm.widgets.waveformplot.WaveformCurveItem.setData')
@mock.patch('pyqtgraph.BarGraphItem.setOpts')
def test_redraw_plot(mocked_set_opts, mocked_set_data, qtbot, monkeypatch):
    """ Test redrawing a waveform plot using both a line and a bar graph """

    # Create a waveform plot and add two data items to it, one to be rendered as a line and one as a bar graph
    waveform_plot = PyDMWaveformPlot()
    line_item = WaveformCurveItem()
    bar_item = WaveformCurveItem(plot_style='Bar')
    bar_item.bar_graph_item = BarGraphItem(x=[], height=[], width=1.0)
    waveform_plot.addCurve(line_item)
    waveform_plot.addCurve(bar_item)

    # Setup some mock data for our data items
    line_item.receiveXWaveform(np.array([1, 5, 10], dtype=float))
    line_item.receiveYWaveform(np.array([10, 15, 12], dtype=float))
    bar_item.receiveXWaveform(np.array([0.5, 1, 1.5, 2, 10, 11], dtype=float))
    bar_item.receiveYWaveform(np.array([45, 50, 52, 40, 24, 30], dtype=float))

    waveform_plot.set_needs_redraw()

    # Simulate a redraw of the plot
    waveform_plot.redrawPlot()

    # The line item should result in a call to set data displaying all available data points as defined above
    assert np.array_equal(mocked_set_data.call_args_list[2][1]['x'], np.array([1, 5, 10]))
    assert np.array_equal(mocked_set_data.call_args_list[2][1]['y'], np.array([10, 15, 12]))

    # As should the bar item, using the set_opts call instead of setData
    assert np.array_equal(mocked_set_opts.call_args_list[1][1]['x'], np.array([0.5, 1, 1.5, 2, 10, 11]))
    assert np.array_equal(mocked_set_opts.call_args_list[1][1]['height'], np.array([45, 50, 52, 40, 24, 30]))

    # After a call to redraw, the plot returns to this state until more data arrives
    assert not waveform_plot._needs_redraw


def test_mismatched_shapes(qtbot):
    """ Test that the logic around waveforms with differing lengths works as expected """

    # Create a waveform plot and add data whose waveform components do not share the same length
    waveform_plot = PyDMWaveformPlot()
    data_item_1 = WaveformCurveItem()

    # Start with the basic case, both waveforms share the same length
    data_item_1.receiveXWaveform(np.array([1, 5, 10, 15], dtype=float))
    data_item_1.receiveYWaveform(np.array([10, 11, 12, 13], dtype=float))
    data_item_1.redrawCurve()

    # The data should remain exactly as is, no truncation required
    assert np.array_equal(data_item_1.x_waveform, np.array([1, 5, 10, 15]))
    assert np.array_equal(data_item_1.y_waveform, np.array([10, 11, 12, 13]))

    data_item_1.receiveXWaveform(np.array([1, 5, 10], dtype=float))
    data_item_1.receiveYWaveform(np.array([10, 11, 12, 13], dtype=float))
    data_item_1.redrawCurve()

    # Now the y-waveform was longer than the x one, so it gets truncated to match the length of x
    assert np.array_equal(data_item_1.x_waveform, np.array([1, 5, 10]))
    assert np.array_equal(data_item_1.y_waveform, np.array([10, 11, 12]))

    data_item_1.receiveXWaveform(np.array([1, 5, 10, 15], dtype=float))
    data_item_1.receiveYWaveform(np.array([10, 11, 12], dtype=float))
    data_item_1.redrawCurve()

    # Opposite case: the x-waveform was longer than the y one, so it gets truncated to match the length of y
    assert np.array_equal(data_item_1.x_waveform, np.array([1, 5, 10]))
    assert np.array_equal(data_item_1.y_waveform, np.array([10, 11, 12]))



from .qtplugin_base import PyDMDesignerPlugin, WidgetCategory
from .tab_bar import PyDMTabWidget


class TabWidgetPlugin(PyDMDesignerPlugin):
    """TabWidgetPlugin needs a custom plugin so that it can
    populate itself with an initial tab."""
    TabClass = PyDMTabWidget

    def __init__(self, extensions=None):
        super(TabWidgetPlugin, self).__init__(self.TabClass,
                                              group=WidgetCategory.CONTAINER,
                                              extensions=extensions)

    def domXml(self):
        """
        XML Description of the widget's properties.
        """
        return ("<widget class=\"{0}\" name=\"{0}\">\n"
                " <property name=\"toolTip\" >\n"
                "  <string>{1}</string>\n"
                " </property>\n"
                " <property name=\"whatsThis\" >\n"
                "  <string>{2}</string>\n"
                " </property>\n"
                "<property name=\"alarmChannels\">\n"
                "<stringlist>\n"
                "  <string></string>\n"
                "</stringlist>\n"
                "</property>\n"
                "<widget class=\"QWidget\" name=\"tab\">\n"
                " <attribute name=\"title\">\n"
                "  <string>Page 1</string>\n"
                " </attribute>\n"
                "</widget>\n"
                "</widget>"
                ).format(self.name(), self.toolTip(), self.whatsThis())



import os
import sys
import imp
import uuid
import inspect
import logging
import platform
import functools
import collections

from qtpy.QtWidgets import QMenu

from .tools import *
from ..utilities import path_info

logger = logging.getLogger(__name__)
ext_tools = {}


def install_external_tool(tool):
    """
    Install an External Tool at the PyDMApplication and add it to the
    main window Tools menu.

    Parameters
    ----------
    tool : str or pydm.tools.ExternalTool
        The full path to a file containing a ExternalTool definition
        or an Instance of an ExternalTool.
    """
    global ext_tools

    try:
        if isinstance(tool, str):
            base_dir, _, _ = path_info(tool)
            sys.path.append(base_dir)
            temp_name = str(uuid.uuid4())

            module = imp.load_source(temp_name, tool)
            classes = [obj for _, obj in inspect.getmembers(module)
                       if inspect.isclass(obj) and issubclass(obj,
                                                              ExternalTool) and obj != ExternalTool]
            if len(classes) == 0:
                raise ValueError(
                    "Invalid File Format. {} has no class inheriting from ExternalTool. Nothing to open at this time.".format(
                        tool))
            obj = [c() for c in classes]
        elif isinstance(tool, ExternalTool):
            # The actual tool to be installed...
            obj = [tool]
        else:
            raise ValueError(
                "Invalid argument for parameter 'tool'. String or ExternalTool expected.")

        for o in obj:
            if o.group is not None and o.group != "":
                if o.group not in ext_tools:
                    ext_tools[o.group] = dict()
                ext_tools[o.group][o.name] = o
            else:
                ext_tools[o.name] = o

        ext_tools = collections.OrderedDict(sorted(ext_tools.items()))
        for k in ext_tools.keys():
            if isinstance(ext_tools[k], dict):
                ext_tools[k] = collections.OrderedDict(
                    sorted(ext_tools[k].items()))
    except Exception as e:
        logger.exception("Failed to load External Tool: %s." % tool)


def assemble_tools_menu(parent_menu, clear_menu=False, widget_only=False,
                        widget=None, **kwargs):
    """
    Assemble the Tools menu for a given parent menu.

    Parameters
    ----------
    parent_menu : QMenu
        The main menu item to hold the tools menu tree.
    clear_menu : bool
        Whether of not we should clear the menu before adding the tools.
    widget_only : bool
        Whether or not generate only the menu for widgets compatible
        tools. This should be True when creating the menu for the
        PyDMWidgets and False for most of the other cases.
    widget: QWidget
        The widget for which the menu is being assembled. This allow for the
        tools to filter if they are compatible or not with the widget based
        on properties, types and etc. Default is None which means all tools
        will be assembled.
    kwargs : dict
        Parameters sent directly to the `call` method of the ExternalTool
        instance. In general this dict is composed by `channels` which
        is a list and `sender` which is a QWidget.

    """
    def assemble_action(menu, tool_obj):
        if tool_obj.icon is not None:
            action = menu.addAction(tool_obj.name)
            action.setIcon(tool_obj.icon)
        else:
            action = menu.addAction(tool_obj.name)
        action.triggered.connect(functools.partial(tool_obj.call, **kwargs))

    load_external_tools()

    if clear_menu:
        parent_menu.clear()
    else:
        parent_menu.addSeparator()

    for k, v in ext_tools.items():
        if isinstance(v, dict):
            m = QMenu(k, parent=parent_menu)
            should_create_menu = False
            for _, t in v.items():
                if widget_only:
                    if not t.use_with_widgets:
                        continue
                    if widget is not None and not t.is_compatible_with(widget):
                        logger.debug('Skipping tool {} as it is incompatible with widget {}.'.format(t.name, widget))
                        continue
                else:
                    if not t.use_without_widget:
                        continue
                assemble_action(m, t)
                should_create_menu = True
            if should_create_menu:
                parent_menu.addMenu(m)
        else:
            if widget_only and not v.use_with_widgets:
                continue
            assemble_action(parent_menu, v)


def load_external_tools():
    """
    Loads all the external tools available at the given
    `PYDM_TOOLS_PATH` environment variable and subfolders that
    follows the `*_tool.py` and have classes that inherits from
    the `pydm.tools.ExternalTool` class.
    """
    if not ext_tools:
        EXT_TOOLS_TOKEN = "_tool.py"
        path = os.getenv("PYDM_TOOLS_PATH", None)

        if path is not None:
            logger.debug("Looking for external tools at: {}".format(path))
            if platform.system() == "Windows":
                locations = path.split(";")
            else:
                locations = path.split(":")
            for loc in locations:
                for root, _, files in os.walk(loc):
                    for name in files:
                        if name.endswith(EXT_TOOLS_TOKEN):
                            install_external_tool(os.path.join(root, name))
        else:
            logger.debug(
                "External Tools not loaded. No External Tools Path specified.")



import pytest
from unittest import mock
from pyqtgraph import AxisItem, ViewBox
from qtpy.QtWidgets import QGraphicsScene
from ...widgets.multi_axis_plot import MultiAxisPlot


@pytest.fixture
def sample_plot(qtbot, monkeypatch):
    """ Set up a plot with a couple of test axis items for use in test cases below """
    plot = MultiAxisPlot()
    scene = QGraphicsScene()
    monkeypatch.setattr(plot, 'scene', lambda: scene)  # Have the plot return a dummy scene when needed

    # Create two axes, one with a fixed range and one that should auto-range (adjust to always display all data)
    fixed_range_axis = AxisItem('left')
    plot.addAxis(fixed_range_axis, 'Fixed Range Axis', enableAutoRangeY=False, minRange=-5.0, maxRange=7.5)
    auto_range_axis = AxisItem('left')
    plot.addAxis(auto_range_axis, 'Auto Range Axis', enableAutoRangeY=True)

    return plot


def test_original_ranges_set_correctly(sample_plot):
    """ Check that the original ranges of x and y axes are preserved when adding them into the plot """

    # The axis with a set range should have that preserved
    assert sample_plot.axesOriginalRanges['Fixed Range Axis'][0] == -5.0
    assert sample_plot.axesOriginalRanges['Fixed Range Axis'][1] == 7.5

    # And the axis set to auto-range should be set to None indicating there is no fixed range to preserve
    assert sample_plot.axesOriginalRanges['Auto Range Axis'][0] is None
    assert sample_plot.axesOriginalRanges['Auto Range Axis'][1] is None


@mock.patch('pyqtgraph.ViewBox.enableAutoRange')
@mock.patch('pyqtgraph.PlotItem.setXRange')
@mock.patch('pyqtgraph.ViewBox.setYRange')
def test_restore_axis_ranges(mocked_y_range, mocked_x_range, mocked_auto_range, sample_plot):
    """
    Verify that when axis ranges have been changed from their defaults, calling restore will set them back to
    their original values
    """

    # First let's set a starting range on the bottom axis as well as those set in the fixture
    sample_plot.axesOriginalRanges['bottom'] = (0.0, 10.0)

    sample_plot.restoreAxisRanges()

    # After the call to restore axis ranges, ensure the view box calls are made as expected. (Don't want to test
    # view box functionality here as that is very involved and covered in pyqtgraph itself)
    mocked_auto_range.assert_any_call(axis=ViewBox.YAxis, enable=True)  # Auto-range enabled for our auto-range axis
    mocked_y_range.assert_called_once_with(-5.0, 7.5)  # Fixed range restored for the other axis
    mocked_x_range.assert_called_once_with(0.0, 10.0, padding=0)  # X-axis restored to the values specified above



import pytest

from qtpy.QtCore import Qt, QSize

from ...widgets.enum_button import PyDMEnumButton, WidgetType, class_for_type
from ... import data_plugins


def test_construct(qtbot):
    """
    Test the construction of the widget.

    Expectations:
    All the default values are properly set.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    """
    widget = PyDMEnumButton()
    qtbot.addWidget(widget)

    assert widget._has_enums is False
    assert widget.orientation == Qt.Vertical
    assert widget.widgetType == WidgetType.PushButton
    assert widget.minimumSizeHint() == QSize(50, 100)


@pytest.mark.parametrize("widget_type", [
    WidgetType.PushButton,
    WidgetType.RadioButton
])
def test_widget_type(qtbot, widget_type):
    """
    Test the widget for a change in the widget type.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    widget_type : WidgetType
        The type of widget to use.
    """
    widget = PyDMEnumButton()
    qtbot.addWidget(widget)

    assert widget.widgetType == WidgetType.PushButton
    assert isinstance(widget._widgets[0], class_for_type[WidgetType.PushButton])

    widget.widgetType = widget_type
    assert widget.widgetType == widget_type
    assert isinstance(widget._widgets[0], class_for_type[widget_type])


@pytest.mark.parametrize("orientation", [
    Qt.Horizontal,
    Qt.Vertical
])
def test_widget_orientation(qtbot, orientation):
    """
    Test the widget for a change in the orientation.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    orientation : int
        One of Qt.Vertical or Qt.Horizontal
    """
    widget = PyDMEnumButton()
    qtbot.addWidget(widget)

    assert widget.orientation == Qt.Vertical

    widget.orientation = orientation
    assert widget.orientation == orientation

    if orientation == Qt.Horizontal:
        row = 0
        col = 1
    else:
        row = 1
        col = 0

    item = widget.layout().itemAtPosition(row, col)
    assert item is not None
    w = item.widget()
    qtbot.addWidget(w)
    assert w is not None
    assert isinstance(w, class_for_type[widget.widgetType])


@pytest.mark.parametrize("connected, write_access, has_enum, is_app_read_only", [
    (True, True, True, True),
    (True, True, True, False),

    (True, True, False, True),
    (True, True, False, False),

    (True, False, False, True),
    (True, False, False, False),

    (True, False, True, True),
    (True, False, True, False),

    (False, True, True, True),
    (False, True, True, False),

    (False, False, True, True),
    (False, False, True, False),

    (False, True, False, True),
    (False, True, False, False),

    (False, False, False, True),
    (False, False, False, False),
])
def test_check_enable_state(qtbot, connected, write_access, has_enum,
                            is_app_read_only):
    """
    Test the tooltip generated depending on the channel connection, write access,
    whether the widget has enum strings,
    and whether the app is read-only.

    Expectations:
    1. If the data channel is disconnected, the widget's tooltip will display "PV is disconnected"
    2. If the data channel is connected, but it has no write access:
        a. If the app is read-only, the tooltip will read  "Running PyDM on Read-Only mode."
        b. If the app is not read-only, the tooltip will read "Access denied by Channel Access Security."
    3. If the widget does not have any enum strings, the tooltip will display "Enums not available".

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    connected : bool
        True if the channel is connected; False otherwise
    write_access : bool
        True if the widget has write access to the channel; False otherwise
    has_enum: bool
        True if the widget has enum strings populated; False if the widget contains no enum strings (empty of choices)
    is_app_read_only : bool
        True if the PyDM app is read-only; False otherwise
    """
    widget = PyDMEnumButton()
    qtbot.addWidget(widget)

    original_tooltip = "Original Tooltip"
    widget.setToolTip(original_tooltip)

    widget.write_access_changed(write_access)

    widget.connection_changed(connected)

    if has_enum:
        widget.enum_strings_changed(("START", "STOP", "PAUSE"))
        assert widget._has_enums

    data_plugins.set_read_only(is_app_read_only)

    widget.check_enable_state()

    actual_tooltip = widget.toolTip()
    if not widget._connected:
        assert "Channel is disconnected." in actual_tooltip
    elif not write_access:
        if data_plugins.is_read_only():
            assert "Running PyDM on Read-Only mode." in actual_tooltip
        else:
            assert "Access denied by Channel Access Security." in actual_tooltip
    elif not widget._has_enums:
        assert "Enums not available" in actual_tooltip


def test_items_is_not_none(qtbot):
    """
    Test that items property is not None, as QStringList definition meant for Qt Designer cannot handle that.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    widget = PyDMEnumButton()
    qtbot.addWidget(widget)
    assert widget.items is not None
    assert isinstance(widget.items, list)


def test_items_defines_button_group(qtbot):
    """
    Test that defining items via the property actually alters buttons.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    widget = PyDMEnumButton()
    qtbot.addWidget(widget)
    widget.items = ["PLAY", "PAUSE"]
    assert len(widget._btn_group.buttons()) == 2
    assert widget._btn_group.button(0).text() == "PLAY"
    assert widget._btn_group.button(1).text() == "PAUSE"
    widget.items = ["STOP"]
    assert len(widget._btn_group.buttons()) == 1
    assert widget._btn_group.button(0).text() == "STOP"


def test_enum_strings_signal_alters_items_prop(qtbot, signals):
    """
    Test that receiving new enum_strings from the CS overrides predefined button items.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    """
    widget = PyDMEnumButton()
    qtbot.addWidget(widget)
    signals.enum_strings_signal[tuple].connect(widget.enumStringsChanged)
    widget.items = ["PLAY", "PAUSE"]
    assert len(widget._btn_group.buttons()) == 2
    assert widget._btn_group.button(0).text() == "PLAY"
    assert widget._btn_group.button(1).text() == "PAUSE"
    signals.enum_strings_signal[tuple].emit(("STOP", ))
    assert len(widget._btn_group.buttons()) == 1
    assert widget._btn_group.button(0).text() == "STOP"


def test_send_receive_value(qtbot, signals):
    """
    Test the widget for round-trip data transfer.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    """
    widget = PyDMEnumButton()
    qtbot.addWidget(widget)

    signals.write_access_signal[bool].connect(widget.writeAccessChanged)
    signals.connection_state_signal[bool].connect(widget.connectionStateChanged)
    signals.new_value_signal[int].connect(widget.channelValueChanged)
    signals.enum_strings_signal[tuple].connect(widget.enumStringsChanged)

    widget.send_value_signal[int].connect(signals.receiveValue)

    signals.write_access_signal[bool].emit(True)
    signals.connection_state_signal[bool].emit(True)
    signals.enum_strings_signal[tuple].emit(("START", "STOP", "PAUSE"))

    signals.new_value_signal[int].emit(1)
    assert widget.value == 1
    assert widget._widgets[1].isChecked()
    widget._widgets[2].click()
    assert not widget._widgets[1].isChecked()
    assert widget._widgets[2].isChecked()
    assert signals.value == 2
